"""
This is adapted from the PyObjC PythonBrowser demo.
I beleive that demo was written by Just van Rossum.
"""

import AppKit
from objc import python_method
from operator import getitem, setitem

import inspect

from vanilla.vanillaBase import VanillaBaseObject
from vanilla.nsSubclasses import getNSSubclass

import warnings
warnings.filterwarnings("ignore", category=UserWarning)

TYPE_COLUMN_MAP = {
    "list" : "List",
    "dict" : "Dict",
    "NoneType" : "None",
    "instance" : "",
    "int" : "Integer",
    "float" : "Float",
    "str" : "String",
    "instancemethod" : "Method"
}

class ObjectBrowser(VanillaBaseObject):

    """
    An object browser.

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing the position and
    size of the browser.

    **obj** The object to be displayed.
    """

    def __init__(self, posSize, obj):
        self._model = PythonBrowserModel.alloc().initWithObject_(obj)

        self._posSize = posSize

        self._nsObject = getNSSubclass("NSScrollView")(self)
        self._nsObject.setAutohidesScrollers_(True)
        self._nsObject.setHasHorizontalScroller_(True)
        self._nsObject.setHasVerticalScroller_(True)
        self._nsObject.setBorderType_(AppKit.NSBezelBorder)
        self._nsObject.setDrawsBackground_(True)

        self._outlineView = getNSSubclass("NSOutlineView")(self)
        self._outlineView.setFrame_(((0, 0), (100, 100)))
        self._outlineView.setUsesAlternatingRowBackgroundColors_(True)
        self._outlineView.setAllowsColumnResizing_(True)
        self._outlineView.setRowHeight_(17.0)

        self._outlineView.setColumnAutoresizingStyle_(AppKit.NSTableViewUniformColumnAutoresizingStyle)
        columns = [
            ("name", "Name"),
            ("type", "Type"),
            ("value", "Value"),
            ("arguments", "Arguments")
        ]
        for key, title in columns:
            column = AppKit.NSTableColumn.alloc().initWithIdentifier_(key)
            column.setResizingMask_(AppKit.NSTableColumnAutoresizingMask | AppKit.NSTableColumnUserResizingMask)
            column.headerCell().setTitle_(title)
            dataCell = column.dataCell()
            dataCell.setDrawsBackground_(False)
            dataCell.setStringValue_("")  # cells have weird default values
            column.setEditable_(False)
            self._outlineView.addTableColumn_(column)
            if key == "name":
                self._outlineView.setOutlineTableColumn_(column)

        self._outlineView.setDataSource_(self._model)
        self._outlineView.setDelegate_(self._model)

        self._nsObject.setDocumentView_(self._outlineView)
        self._setAutosizingFromPosSize(posSize)

    def getNSScrollView(self):
        return self._nsObject

    def getNSOutlineView(self):
        return self._outlineView


class PythonBrowserModel(AppKit.NSObject):

    """This is a delegate as well as a data source for NSOutlineViews."""

    def initWithObject_(self, obj):
        self = self.init()
        self.setObject_(obj)
        return self

    def setObject_(self, obj):
        self.root = PythonItem("<root>", obj, None, None)

    # NSOutlineViewDataSource  methods

    def outlineView_numberOfChildrenOfItem_(self, view, item):
        if item is None:
            item = self.root
        return len(item)

    def outlineView_child_ofItem_(self, view, child, item):
        if item is None:
            item = self.root
        return item.getChild(child)

    def outlineView_isItemExpandable_(self, view, item):
        if item is None:
            item = self.root
        return item.isExpandable()

    def outlineView_objectValueForTableColumn_byItem_(self, view, col, item):
        if item is None:
            item = self.root
        identifier = col.identifier()
        value = getattr(item, identifier)
        # filter the type values
        if identifier == "type":
            value = TYPE_COLUMN_MAP.get(value, value)
        return value

    def outlineView_shouldEditTableColumn_item_(self, view, col, item):
        return False

    def outlineView_toolTipForCell_rect_tableColumn_item_mouseLocation_(self, view, cell, rect, col, item, location):
        ## addig a tooltip, use the __doc__ from the object
        return item.getDoc(), rect

# objects of these types are not eligable for expansion in the outline view
SIMPLE_TYPES = (str, int, float, complex)

def getChilderen(root):
    childeren = []

    for name, obj in inspect.getmembers(root):
        ## ignore private methods and attributes
        if name.startswith("_"):
            continue
        ## ignore imported modules
        #elif inspect.ismodule(obj):
        #    continue
        ## ignore methods and attributed usind in pyobjc
        elif name.startswith("pyobjc_"):
            continue
        ## ignore methods and attributed usind in pyobjc
        elif type(obj).__name__ in ["native_selector"]:
            continue
        childeren.append(name)

    return childeren


def getArguments(obj):
    """
    Return all arguments for a method of function
    and leave 'self' out.
    """
    try:
        arguments = inspect.formatargspec(*inspect.getargspec(obj))
    except TypeError:
        arguments = ""
    return arguments.replace("self, ", "").replace("self", "")

class PythonItem(AppKit.NSObject):

    """Wrapper class for items to be displayed in the outline view."""

    # We keep references to all child items (once created). This is
    # neccesary because NSOutlineView holds on to PythonItem instances
    # without retaining them. If we don"t make sure they don"t get
    # garbage collected, the app will crash. For the same reason this
    # class _must_ derive from NSObject, since otherwise autoreleased
    # proxies will be fed to NSOutlineView, which will go away too soon.

    def __new__(cls, *args, **kwargs):
        # "Pythonic" constructor
        return cls.alloc().init()

    def __init__(self, name, obj, parent, setvalue, ignoreAppKit=True):
        self.realName = name
        self.name = str(name)
        self.parent = parent
        self.arguments = ""
        self.type = type(obj).__name__
        if obj is None:
            self.value = "None"
        elif not isinstance(obj, SIMPLE_TYPES):
            self.value = ""
        else:
            self.value = obj

        ## in pyOjbc a python_selector should have an attr callable with is actually the method or function
        if self.type == "python_selector" and hasattr(obj, "callable"):
            obj = obj.callable

        self.object = obj
        self.children = []

        self.getters = dict()
        self.setters = dict()
        if isinstance(obj, dict):
            self.children = sorted(obj.keys())
            self._setGetters(self.children, getitem)
            self._setSetters(self.children, setitem)
        elif obj is None or isinstance(obj, SIMPLE_TYPES):
            pass
        elif isinstance(obj, (list, tuple, set)):
            self.children = list(range(len(obj)))
            self._setGetters(self.children, getitem)
            self._setSetters(self.children, setitem)
        elif isinstance(obj, property):
            pass
        elif inspect.ismethod(obj):
            self.arguments = getArguments(obj)
        elif inspect.isfunction(obj):
            self.arguments = getArguments(obj)
        else:
            try:
                l = list(obj)
                self.children = list(range(len(l)))
                self._setGetters(self.children, getitem)
                self._setSetters(self.children, setitem)
            except:
                pass

            try:
                d = dict(obj)
                self.children = sorted(d.keys())
                self._setGetters(self.children, getitem)
                self._setSetters(self.children, setitem)
            except:
                pass
            children = getChilderen(obj)
            self._setGetters(children, getattr)
            self._setSetters(children, setattr)
            self.children += children

            if inspect.isclass(obj) and hasattr(obj, "__init__"):
                self.arguments = getArguments(getattr(obj, "__init__"))

        if ignoreAppKit:
            self.children = [child for child in self.children if not (isinstance(child, str) and hasattr(AppKit, child))]

        self._childRefs = {}

    @python_method
    def _setSetters(self, names, callback):
        for name in names:
            self.setters[name] = callback

    @python_method
    def _setGetters(self, names, callback):
        for name in names:
            self.getters[name] = callback

    def isExpandable(self):
        return bool(self.children)

    @python_method
    def getChild(self, child):
        if child in self._childRefs:
            return self._childRefs[child]

        name = self.children[child]
        getter = self.getters.get(name)
        setter = self.setters.get(name)
        obj = getter(self.object, name)

        childObj = self.__class__(name, obj, self.object, setter)
        self._childRefs[child] = childObj
        return childObj

    def getDoc(self):
        doc = inspect.getdoc(self.object)
        if doc:
            return doc
        return None

    def __len__(self):
        return len(self.children)


if __name__ == "__main__":
    import vanilla
    testObject = vanilla

    class TestWindow():
        def __init__(self):
            self.w = vanilla.Window((400, 400), "inspect object browser %s" %testObject, minSize=(100, 100))
            self.w.b = ObjectBrowser((0, 0, -0, -0), testObject)
            self.w.open()


    from vanilla.test.testTools import executeVanillaTest
    executeVanillaTest(TestWindow)










from distutils.core import setup
import py2app
import os

plist = dict(
    CFBundleDocumentTypes = [
    dict(
    	CFBundleTypeExtensions = ["txt"],
    	CFBundleTypeName = "Text",
    	CFBundleTypeRole = "Editor",
    	NSDocumentClass = "TinyTextEditorDocument",
    	LSTypeIsPackage = True,
        ),
        ],
	CFBundleIdentifier = "com.typesupply.TinyTextEditor",
	LSMinimumSystemVersion = "10.3.9",
	CFBundleShortVersionString = "1.0.0",
	CFBundleVersion = "1.0.0",
	)

dataFiles = [
		'Resources/English.lproj',
		]

setup(
	data_files=dataFiles,
	app=[dict(script="TinyTextEditor.py", plist=plist)]
	)



from distutils.core import setup
import py2app
import os

plist = dict(
    CFBundleIdentifier = "com.yourDomain.SimpleApp",
    LSMinimumSystemVersion = "10.4.0",
    CFBundleShortVersionString = "1.0.0",
    CFBundleVersion = "1.0.0",
    )

dataFiles = [
        'Resources/English.lproj',
        ]

setup(
    data_files=dataFiles,
    app=[dict(script="SimpleApp.py", plist=plist)]
    )



from Foundation import NSMaxXEdge, NSMaxYEdge, NSMinXEdge, NSMinYEdge
from AppKit import NSDrawer
from vanilla.vanillaBase import VanillaBaseObject, _breakCycles


_drawerEdgeMap = {
    "left": NSMinXEdge,
    "right": NSMaxXEdge,
    "top": NSMaxYEdge,
    "bottom": NSMinYEdge,
}


class Drawer(VanillaBaseObject):

    """
    A drawer attached to a window. Drawers are capable of containing controls.

    .. image:: /_images/Drawer.png

    .. warning:: Drawers are deprecated and should not be used in modern macOS apps.

    To add a control to a drawer, simply set it as an attribute of the drawer.

    ::

        from vanilla import Window, Button, Drawer, TextBox

        class DrawerDemo:

            def __init__(self):
                self.w = Window((200, 200))
                self.w.button = Button((10, 10, -10, 20), "Toggle Drawer",
                                    callback=self.toggleDrawer)
                self.d = Drawer((100, 150), self.w)
                self.d.textBox = TextBox((10, 10, -10, -10),
                                    "This is a drawer.")
                self.w.open()
                self.d.open()

            def toggleDrawer(self, sender):
                self.d.toggle()

        DrawerDemo()

    No special naming is required for the attributes. However, each attribute
    must have a unique name.

    **size** Tuple of form *(width, height)* representing the size of the drawer.

    **parentWindow** The window that the drawer should be attached to.

    **minSize** Tuple of form *(width, height)* representing the minimum size
    of the drawer.

    **maxSize** Tuple of form *(width, height)* representing the maximum size
    of the drawer.

    **preferredEdge** The preferred edge of the window that the drawer should be
    attached to. If the drawer cannot be opened on the preferred edge, it will
    be opened on the opposite edge. The options are:

    +----------+
    | "left"   |
    +----------+
    | "right"  |
    +----------+
    | "top"    |
    +----------+
    | "bottom" |
    +----------+

    **forceEdge** Boolean representing if the drawer should *always* be opened
    on the preferred edge.

    **leadingOffset** Distance between the top or left edge of the drawer
    and the parent window.

    **trailingOffset** Distance between the bottom or right edge of the drawer
    and the parent window.
    """

    nsDrawerClass = NSDrawer

    def __init__(self, size, parentWindow, minSize=None, maxSize=None,
            preferredEdge="left", forceEdge=False, leadingOffset=20, trailingOffset=20):
        from vanilla.vanillaWindows import Window
        self._preferredEdge = preferredEdge
        self._forceEdge = forceEdge
        drawer = self._nsObject = self.nsDrawerClass.alloc().initWithContentSize_preferredEdge_(
                size, _drawerEdgeMap[preferredEdge])
        drawer.setLeadingOffset_(leadingOffset)
        drawer.setTrailingOffset_(trailingOffset)
        if minSize:
            drawer.setMinContentSize_(minSize)
        if maxSize:
            drawer.setMaxContentSize_(maxSize)
        if isinstance(parentWindow, Window):
            parentWindow = parentWindow._window
        drawer.setParentWindow_(parentWindow)

    def getNSDrawer(self):
        """
        Return the `NSDrawer`_ that this object wraps.

        .. _NSDrawer: https://developer.apple.com/documentation/appkit/nsdrawer?language=objc
        """
        return self._nsObject

    def _getContentView(self):
        return self._nsObject.contentView()

    def _breakCycles(self):
        super(Drawer, self)._breakCycles()
        view = self._getContentView()
        if view is not None:
            _breakCycles(view)

    def open(self):
        """
        Open the drawer.
        """
        if self._forceEdge:
            self._nsObject.openOnEdge_(_drawerEdgeMap[self._preferredEdge])
        else:
            self._nsObject.open()

    def close(self):
        """
        Close the drawer.
        """
        self._nsObject.close()

    def toggle(self):
        """
        Open the drawer if it is closed. Close it if it is open.
        """
        self._nsObject.toggle_(self)

    def isOpen(self):
        """
        Return *True* if the drawer is open, *False* if it is closed.
        """
        return self._nsObject.isOpen()



from AppKit import NSSmallSquareBezelStyle
from vanilla.vanillaButton import ImageButton


class GradientButton(ImageButton):

    """
    An image button that initiates an immediate action related to a view.

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing
    the position and size of the button.

    **title** The text to be displayed on the button. Pass *None* if no title is desired.

    **bordered** Boolean representing if the button should be bordered.

    **imagePath** A file path to an image.

    **imageNamed** The name of an image already loaded as a `NSImage`_ by the application.

    **imageObject** A `NSImage`_ object.

    .. note:: Only one of *imagePath*, *imageNamed*, *imageObject* should be set.

    **imagePosition** The position of the image relative to the title.
    The options are:

    +----------+
    | "top"    |
    +----------+
    | "bottom" |
    +----------+
    | "left"   |
    +----------+
    | "right"  |
    +----------+

    **callback** The method to be called when the user presses the button.

    **sizeStyle** A string representing the desired size style of the button.
    The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+

    .. _NSImage: https://developer.apple.com/documentation/appkit/nsimage?language=objc
    """

    nsBezelStyle = NSSmallSquareBezelStyle



import AppKit
from vanilla import VanillaBaseObject

columnPlacements = dict(
    leading=AppKit.NSGridCellPlacementLeading,
    center=AppKit.NSGridCellPlacementCenter,
    trailing=AppKit.NSGridCellPlacementTrailing,
    fill=AppKit.NSGridCellPlacementFill
)
rowPlacements = dict(
    top=AppKit.NSGridCellPlacementTop,
    center=AppKit.NSGridCellPlacementCenter,
    bottom=AppKit.NSGridCellPlacementBottom,
    fill=AppKit.NSGridCellPlacementFill
)
rowAlignments = dict(
    firstBaseline=AppKit.NSGridRowAlignmentFirstBaseline,
    lastBaseline=AppKit.NSGridRowAlignmentLastBaseline,
    none=AppKit.NSGridRowAlignmentNone
)

class GridView(VanillaBaseObject):

    """
    A view that allows the placement of other views within a grid.

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing
    the position and size of the grid view.

    **contents** The contents to display within the grid. See below for structure.

    **columnWidth** The width for columns.

    **columnSpacing** The amount of spacing between columns.

    **columnPadding** The (left, right) padding for columns.

    **columnPlacement** The horizontal placement of content within columns. Options:

    * "leading"
    * "center"
    * "trailing"
    * "fill"

    **rowHeight** The height for rows.

    **rowSpacing** The amount of spacing between rows.

    **rowPadding** The (top, bottom) padding for rows.

    **rowPlacement** The vertical placement of content within rows. Options:

    * "top"
    * "center"
    * "bottom"
    * "fill"

    **rowAlignment** The alignment of the row. Options:

    * "firstBaseline"
    * "lastBaseline"
    * "none"

    **columnDescriptions** An optional list of dictionaries
    defining specific attributes for the columns. Options:

    * "width"
    * "columnPadding"
    * "columnPlacement"


    **Contents Definition Structure**

    Contents are defined as with a list of row definitions.
    A row definition is a list of cell definitions or a
    dictionary with this structure:

    * **cells** A list of cell definitions.
    * **rowHeight** (optional) A height for the row that overrides
      the GridView level row height.
    * **rowPadding** (optional) A (top, bottom) padding definition
      for the row that overrides the GridView level row padding.
    * **rowPlacement** (optional) A placement for the row that
      overrides the GridView level row placement.
    * **rowAlignment** (optional) An alignment for the row that
      overrides the GridView level row placement.

    Cells are defined with either a Vanilla object, a NSView
    (or NSView subclass) object, None, or a dictionary with
    this structure:

    * **view** A Vanilla object or NSView (or NSView subclass) object.
    * **width** (optional) A width to apply to the view.
    * **height** (optional) A height to apply to the view.
    * **columnPlacement** (optional) A horizontal placement for the
      cell that overrides the row level or GridView level placement.
    * **rowPlacement** (optional) A vertical placement for the cell
      that overrides the row level or GridView level placement.
    * **rowAlignment** (optional) A row alignment for the cell that
      overrides the row level or GridView level alignment.

    If a cell is defined as None, the cell will be merged with the
    first cell directly above that has content.
    """

    nsGridViewClass = AppKit.NSGridView

    def __init__(self,
            posSize,
            contents,
            columnWidth=None,
            columnSpacing=0,
            columnPadding=(0, 0),
            columnPlacement="leading",
            rowHeight=None,
            rowSpacing=0,
            rowPadding=(0, 0),
            rowPlacement="top",
            rowAlignment="firstBaseline",
            columnDescriptions=None
        ):
        if columnDescriptions is None:
            sample = contents[0]
            if isinstance(sample, dict):
                count = len(sample["cells"])
            else:
                count = len(sample)
            columnDescriptions = [{} for i in range(count)]
        self._setupView(self.nsGridViewClass, posSize)
        gridView = self.getNSGridView()
        gridView.setColumnSpacing_(columnSpacing)
        gridView.setRowSpacing_(rowSpacing)
        gridView.setXPlacement_(columnPlacements[columnPlacement])
        gridView.setYPlacement_(rowPlacements[rowPlacement])
        gridView.setRowAlignment_(rowAlignments[rowAlignment])
        self._columnWidth = columnWidth
        self._columnPadding = columnPadding
        self._rowHeight = rowHeight
        self._rowPadding = rowPadding
        self._buildColumns(columnDescriptions)
        self._buildRows(contents)

    def getNSGridView(self):
        return self._getContentView()

    # Input Normalizing

    def _normalizeRows(self, rows):
        rows = [
            self._normalizeRow(row)
            for row in rows
        ]
        return rows

    def _normalizeRow(self, row):
        if not isinstance(row, dict):
            row = dict(cells=row)
        if "height" not in row:
            row["height"] = self._rowHeight
        if "rowPadding" not in row:
            row["rowPadding"] = self._rowPadding
        row["cells"] = self._normalizeCells(row["cells"])
        return row

    def _normalizeCells(self, cells):
        cells = [
            self._normalizeCell(cell)
            for cell in cells
        ]
        return cells

    def _normalizeCell(self, cell):
        if cell is None:
            return None
        if not isinstance(cell, dict):
            cell = dict(view=cell)
        return cell

    # Building

    def _buildColumns(self, columnDescriptions):
        gridView = self.getNSGridView()
        for columnDescription in columnDescriptions:
            column = gridView.addColumnWithViews_([])
            self._setColumnAttributes(column, columnDescription)

    def _setColumnAttributes(self, column, columnDescription):
        width = columnDescription.get("width", self._columnWidth)
        columnPadding = columnDescription.get("columnPadding", self._columnPadding)
        columnPlacement = columnDescription.get("columnPlacement")
        if width is not None:
            column.setWidth_(width)
        column.setLeadingPadding_(columnPadding[0])
        column.setTrailingPadding_(columnPadding[1])
        if columnPlacement is not None:
            column.setXPlacement_(columnPlacements[columnPlacement])

    def _populateColumns(self, columns):
        gridView = self.getNSGridView()
        for columnIndex, cells in enumerate(columns):
            column = gridView.columnAtIndex_(columnIndex)
            self._populateColumn(column, cells)

    def _populateColumn(self, column, cells):
        gridView = self.getNSGridView()
        columnIndex = gridView.indexOfColumn_(column)
        for rowIndex, cellData in enumerate(cells):
            self._populateCell(columnIndex, rowIndex, cellData)

    def _buildRows(self, rows):
        gridView = self.getNSGridView()
        rows = self._normalizeRows(rows)
        for rowIndex in range(len(rows)):
            gridView.addRowWithViews_([])
        # set row attributes
        for rowIndex, rowData in enumerate(rows):
            row = gridView.rowAtIndex_(rowIndex)
            self._setRowAttributes(row, rowData)
        # merge cells
        mergers = {}
        for rowIndex, rowData in enumerate(rows):
            rowMergers = [[]]
            for columnIndex, cell in enumerate(rowData["cells"]):
                if cell is None:
                    rowMergers[-1].append(columnIndex)
                else:
                    if rowMergers[-1]:
                        rowMergers.append([])
            if rowMergers[0]:
                mergers[rowIndex] = rowMergers
        for rowIndex, rowMergers in sorted(mergers.items()):
            for merger in rowMergers:
                if not merger:
                    continue
                start = merger[0] - 1
                # can't merge first column with a nonexistent previous column
                if start == -1:
                    continue
                end = merger[-1] + 1
                length = end - start
                gridView.mergeCellsInHorizontalRange_verticalRange_(
                    AppKit.NSMakeRange(start, length),
                    AppKit.NSMakeRange(rowIndex, 1)
                )
        # populate columns
        columns = {}
        for rowData in rows:
            cells = rowData["cells"]
            for columnIndex, view in enumerate(cells):
                if columnIndex not in columns:
                    columns[columnIndex] = []
                columns[columnIndex].append(view)
        columns = [cells for columnIndex, cells in sorted(columns.items())]
        self._populateColumns(columns)

    def _setRowAttributes(self, row, rowData):
        height = rowData["height"]
        rowPadding = rowData["rowPadding"]
        rowPlacement = rowData.get("rowPlacement")
        rowAlignment = rowData.get("rowAlignment")
        if height is not None:
            row.setHeight_(height)
        row.setTopPadding_(rowPadding[0])
        row.setBottomPadding_(rowPadding[1])
        if rowPlacement is not None:
            row.setYPlacement_(rowPlacements[rowPlacement])
        if rowAlignment is not None:
            row.setRowAlignment_(rowAlignments[rowAlignment])

    def _populateRow(self, row, cells):
        gridView = self.getNSGridView()
        rowIndex = gridView.indexOfRow_(row)
        for columnIndex, cellData in enumerate(cells):
            self._populateCell(columnIndex, rowIndex, cellData)

    def _populateCell(self, columnIndex, rowIndex, cellData):
        if cellData is None:
            return
        gridView = self.getNSGridView()
        view = cellData["view"]
        if isinstance(view, VanillaBaseObject):
            view = view._nsObject
        cell = gridView.cellAtColumnIndex_rowIndex_(columnIndex, rowIndex)
        cell.setContentView_(view)
        columnPlacement = cellData.get("columnPlacement")
        rowPlacement = cellData.get("rowPlacement")
        rowAlignment = cellData.get("rowAlignment")
        width = cellData.get("width")
        height = cellData.get("height")
        # special handling and defaults for
        # views without an intrinsic size
        if view.intrinsicContentSize() == (-1, -1):
            w, h = view.fittingSize()
            if width is None:
                if w > 0:
                    width = w
                else:
                    width = gridView.columnAtIndex_(columnIndex).width()
            if height is None:
                if h > 0:
                    height = h
                else:
                    height = gridView.rowAtIndex_(rowIndex).height()
            if rowAlignment is None:
                rowAlignment = "none"
            if columnPlacement is None:
                columnPlacement = "leading"
            if rowPlacement is None:
                rowPlacement = "top"
        if columnPlacement is not None:
            cell.setXPlacement_(columnPlacements[columnPlacement])
        if rowPlacement is not None:
            cell.setYPlacement_(rowPlacements[rowPlacement])
        if rowAlignment is not None:
            cell.setRowAlignment_(rowAlignments[rowAlignment])
        constraints = []
        if width is not None:
            constraint = AppKit.NSLayoutConstraint.constraintWithItem_attribute_relatedBy_toItem_attribute_multiplier_constant_(
                view,
                AppKit.NSLayoutAttributeWidth,
                AppKit.NSLayoutRelationEqual,
                None,
                AppKit.NSLayoutAttributeWidth,
                1.0,
                width
            )
            constraints.append(constraint)
        if height is not None:
            constraint = AppKit.NSLayoutConstraint.constraintWithItem_attribute_relatedBy_toItem_attribute_multiplier_constant_(
                view,
                AppKit.NSLayoutAttributeHeight,
                AppKit.NSLayoutRelationEqual,
                None,
                AppKit.NSLayoutAttributeHeight,
                1.0,
                height
            )
            constraints.append(constraint)
        if constraints:
            cell.setCustomPlacementConstraints_(constraints)

    # -------
    # Columns
    # -------

    def getColumnCount(self):
        """
        Get the number of columns.
        """
        gridView = self.getNSGridView()
        return gridView.numberOfColumns()

    def columnIsVisible(self, index):
        """
        Get the visibility of column at *index*.
        """
        gridView = self.getNSGridView()
        column = gridView.columnAtIndex_(index)
        return not column.isHidden()

    def showColumn(self, index, value):
        """
        Set the visibility of column at *index*.
        """
        gridView = self.getNSGridView()
        column = gridView.columnAtIndex_(index)
        column.setHidden_(not value)

    def appendColumn(self, cells, columnWidth=None, columnPadding=None, columnPlacement=None):
        """
        Append a column and populate it with a list of cells.
        The cells must have the same structure as defined in *__init__*.
        """
        gridView = self.getNSGridView()
        column = gridView.addColumnWithViews_([])
        columnDescription = {}
        if columnWidth is not None:
            columnDescription["columnWidth"] = columnWidth
        if columnPadding is not None:
            columnDescription["columnPadding"] = columnPadding
        if columnPlacement is not None:
            columnDescription["columnPlacement"] = columnPlacement
        self._setColumnAttributes(column, columnDescription)
        cells = self._normalizeCells(cells)
        self._populateColumn(column, cells)

    def insertColumn(self, index, cells, columnWidth=None, columnPadding=None, columnPlacement=None):
        """
        Insert a column at *index* and populate it with a list of cells.
        The cells must have the same structure as defined in *__init__*.
        """
        gridView = self.getNSGridView()
        column = gridView.insertColumnAtIndex_withViews_(index, [])
        columnDescription = {}
        if columnWidth is not None:
            columnDescription["columnWidth"] = columnWidth
        if columnPadding is not None:
            columnDescription["columnPadding"] = columnPadding
        if columnPlacement is not None:
            columnDescription["columnPlacement"] = columnPlacement
        self._setColumnAttributes(column, columnDescription)
        cells = self._normalizeCells(cells)
        self._populateColumn(column, cells)

    def removeColumn(self, index):
        """
        Remove column at *index*.
        """
        gridView = self.getNSGridView()
        gridView.removeColumnAtIndex_(index)

    def moveColumn(self, fromIndex, toIndex):
        """
        Move column at *fromIndex* to *toIndex*.
        """
        gridView = self.getNSGridView()
        gridView.moveColumnAtIndex_toIndex_(fromIndex, toIndex)

    # ----
    # Rows
    # ----

    def getRowCount(self):
        """
        Get the number of rows.
        """
        gridView = self.getNSGridView()
        return gridView.numberOfRows()

    def showRow(self, index, value):
        """
        Set the visibility of row at *index*.
        """
        gridView = self.getNSGridView()
        row = gridView.rowAtIndex_(index)
        row.setHidden_(not value)

    def rowIsVisible(self, index):
        """
        Get the visibility of row at *index*.
        """
        gridView = self.getNSGridView()
        row = gridView.rowAtIndex_(index)
        return not row.isHidden()

    def appendRow(self, cells, rowHeight=None, rowPadding=None, rowPlacement=None, rowAlignment=None):
        """
        Append a row and populate it with a list of cells.
        The cells must have the same structure as defined in *__init__*.

        Merging is not possible with this method.
        """
        gridView = self.getNSGridView()
        rowDescription = dict(cells=cells)
        if rowHeight is not None:
            rowDescription["rowHeight"] = rowHeight
        if rowPadding is not None:
            rowDescription["rowPadding"] = rowPadding
        if rowPlacement is not None:
            rowDescription["rowPlacement"] = rowPlacement
        if rowAlignment is not None:
            rowDescription["rowAlignment"] = rowAlignment
        rowDescription = self._normalizeRow(rowDescription)
        row = gridView.addRowWithViews_([])
        self._setRowAttributes(row, rowDescription)
        self._populateRow(row, rowDescription["cells"])

    def insertRow(self, index, cells, rowHeight=None, rowPadding=None, rowPlacement=None, rowAlignment=None):
        """
        Insert a row at *index* and populate it with a list of cells.
        The cells definition must have the same structure as defined in *__init__*.

        Merging is not possible with this method.
        """
        gridView = self.getNSGridView()
        rowDescription = dict(cells=cells)
        if rowHeight is not None:
            rowDescription["rowHeight"] = rowHeight
        if rowPadding is not None:
            rowDescription["rowPadding"] = rowPadding
        if rowPlacement is not None:
            rowDescription["rowPlacement"] = rowPlacement
        if rowAlignment is not None:
            rowDescription["rowAlignment"] = rowAlignment
        rowDescription = self._normalizeRow(rowDescription)
        row = gridView.insertRowAtIndex_withViews_(index, [])
        self._setRowAttributes(row, rowDescription)
        self._populateRow(row, rowDescription["cells"])

    def removeRow(self, index):
        """
        Remove row at *index*.
        """
        gridView = self.getNSGridView()
        gridView.removeRowAtIndex_(index)

    def moveRow(self, fromIndex, toIndex):
        """
        Move row at *fromIndex* to *toIndex*.
        """
        gridView = self.getNSGridView()
        gridView.moveRowAtIndex_toIndex_(fromIndex, toIndex)



from AppKit import NSPathControl, NSPathStyleStandard, NSColor, NSFocusRingTypeNone, NSURL, NSPathStylePopUp, NSRoundedBezelStyle
import os
from vanilla.vanillaBase import VanillaBaseControl


_pathStylesMap = {
    "standard": NSPathStyleStandard,
    "popUp": NSPathStylePopUp,
}


class PathControl(VanillaBaseControl):

    """
    A path control.

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing the position
    and size of the control. The size of the control should match the appropriate value
    for the given *sizeStyle*.

    +-------------------------+
    | **Standard Dimensions** |
    +=========+===+===========+
    | Regular | H | 22        |
    +---------+---+-----------+
    | Small   | H | 20        |
    +---------+---+-----------+
    | Mini    | H | 18        |
    +---------+---+-----------+

    **url** The url to be displayed in the control.

    **callback** The method to be called when the user presses the control.

    **pathStyle** A string representing the path style. The options are:

    +------------+
    | "standard" |
    +------------+
    | "popUp"    |
    +------------+

    **sizeStyle** A string representing the desired size style of the button. The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+
    """

    nsPathControlClass = NSPathControl

    def __init__(self, posSize, url, callback=None, pathStyle="standard", sizeStyle="regular"):
        self._setupView(self.nsPathControlClass, posSize, callback=callback)
        self._nsObject.setPathStyle_(_pathStylesMap[pathStyle])
        self._setSizeStyle(sizeStyle)
        self._nsObject.setBackgroundColor_(NSColor.clearColor())
        self._nsObject.setFocusRingType_(NSFocusRingTypeNone)
        self._nsObject.cell().setBordered_(True)
        self._nsObject.cell().setBezelStyle_(NSRoundedBezelStyle)
        self.set(url)

    def set(self, url):
        if url is not None:
            url = NSURL.URLWithString_(url)
        self._nsObject.setURL_(url)

    def get(self):
        url = self._nsObject.URL()
        if url is not None:
            return url.path()
        return None

    def getSelected(self):
        path = []
        for item in self._nsObject.pathItems():
            cell = item.pathComponentCell()
            path.append(item.title())
            if cell == self._nsObject.clickedPathComponentCell():
                break
        return os.sep.join(path)



from Foundation import NSObject
from AppKit import NSApplication, NSMenu, NSMenuItem, NSBundle
from PyObjCTools import AppHelper
import asyncio

try:
    from corefoundationasyncio import CoreFoundationEventLoop
    hasCorefoundationasyncio = True
except ImportError:
    hasCorefoundationasyncio = False


class _VanillaMiniApp(NSApplication):
    pass


class _VanillaMiniAppDelegate(NSObject):

    def applicationShouldTerminateAfterLastWindowClosed_(self, notification):
        return True


def executeVanillaTest(cls, nibPath=None, calls=None, **kwargs):
    """
    Execute a Vanilla UI class in a mini application.
    """
    app = _VanillaMiniApp.sharedApplication()
    delegate = _VanillaMiniAppDelegate.alloc().init()
    app.setDelegate_(delegate)

    if nibPath:
        NSBundle.loadNibFile_externalNameTable_withZone_(nibPath, {}, None)
    else:
        mainMenu = NSMenu.alloc().initWithTitle_("Vanilla Test")

        fileMenuItem = NSMenuItem.alloc().initWithTitle_action_keyEquivalent_("File", None, "")
        fileMenu = NSMenu.alloc().initWithTitle_("File")
        fileMenuItem.setSubmenu_(fileMenu)
        mainMenu.addItem_(fileMenuItem)

        editMenuItem = NSMenuItem.alloc().initWithTitle_action_keyEquivalent_("Edit", None, "")
        editMenu = NSMenu.alloc().initWithTitle_("Edit")
        editMenu.addItemWithTitle_action_keyEquivalent_("Cut", "cut:", "x")
        editMenu.addItemWithTitle_action_keyEquivalent_("Copy", "copy:", "c")
        editMenu.addItemWithTitle_action_keyEquivalent_("Paste", "paste:", "v")
        editMenu.addItem_(NSMenuItem.separatorItem())
        editMenu.addItemWithTitle_action_keyEquivalent_("Select All", "selectAll:", "a")
        editMenuItem.setSubmenu_(editMenu)
        mainMenu.addItem_(editMenuItem)

        helpMenuItem = NSMenuItem.alloc().initWithTitle_action_keyEquivalent_("Help", None, "")
        helpMenu = NSMenu.alloc().initWithTitle_("Help")
        helpMenuItem.setSubmenu_(helpMenu)
        mainMenu.addItem_(helpMenuItem)

        app.setMainMenu_(mainMenu)

    if cls is not None:
        cls(**kwargs)

    if calls is not None:
        for call, kwargs in calls:
            call(**kwargs)

    app.activateIgnoringOtherApps_(True)

    if hasCorefoundationasyncio:
        loop = CoreFoundationEventLoop()
        asyncio.set_event_loop(loop)
        try:
            loop.run_forever()
        finally:
            loop.close()
    else:
        AppHelper.runEventLoop()



from Foundation import NSObject
from AppKit import NSApp, NSWindow, NSPanel, NSScreen, NSWindowController, NSToolbar, NSToolbarItem, NSImage, NSNormalWindowLevel, NSFloatingWindowLevel, NSClosableWindowMask, NSMiniaturizableWindowMask, NSResizableWindowMask, NSTexturedBackgroundWindowMask, NSUnifiedTitleAndToolbarWindowMask, NSHUDWindowMask, NSUtilityWindowMask, NSTitledWindowMask, NSBorderlessWindowMask, NSBackingStoreBuffered, NSToolbarFlexibleSpaceItemIdentifier, NSToolbarSpaceItemIdentifier, NSToolbarSeparatorItemIdentifier, NSToolbarCustomizeToolbarItemIdentifier, NSToolbarPrintItemIdentifier, NSToolbarShowFontsItemIdentifier, NSToolbarShowColorsItemIdentifier, NSToolbarDisplayModeDefault, NSToolbarDisplayModeIconAndLabel, NSToolbarDisplayModeIconOnly, NSToolbarDisplayModeLabelOnly, NSToolbarSizeModeDefault, NSToolbarSizeModeRegular, NSToolbarSizeModeSmall
from objc import python_method

from vanilla.vanillaBase import _breakCycles, _calcFrame, _setAttr, _delAttr, _addAutoLayoutRules, _flipFrame, \
        VanillaCallbackWrapper, VanillaError, VanillaWarning, VanillaBaseControl, \
        osVersionCurrent, osVersion10_7, osVersion10_10, osVersion10_16

# PyObjC may not have these constants wrapped,
# so test and fallback if needed.
try:
    from AppKit import NSWindowCollectionBehaviorFullScreenPrimary, NSWindowCollectionBehaviorFullScreenAuxiliary
except ImportError:
    NSWindowCollectionBehaviorFullScreenPrimary = 1 << 7
    NSWindowCollectionBehaviorFullScreenAuxiliary = 1 << 8

try:
    from AppKit import NSWindowTitleVisible, NSWindowTitleHidden
except ImportError:
    NSWindowTitleVisible  = 0
    NSWindowTitleHidden = 1

try:
    # https://developer.apple.com/documentation/appkit/nswindowtoolbarstyle?language=objc
    from AppKit import NSWindowToolbarStyleAutomatic, NSWindowToolbarStyleExpanded, NSWindowToolbarStylePreference, NSWindowToolbarStyleUnified, NSWindowToolbarStyleUnifiedCompact
except ImportError:
    NSWindowToolbarStyleAutomatic = 0
    NSWindowToolbarStyleExpanded = 1
    NSWindowToolbarStylePreference = 2
    NSWindowToolbarStyleUnified = 3
    NSWindowToolbarStyleUnifiedCompact = 4

try:
    from AppKit import NSFullSizeContentViewWindowMask
except ImportError:
    NSFullSizeContentViewWindowMask = 1 << 15


class Window(NSObject):

    """
    A window capable of containing controls.

    .. image:: /_images/Window.png

    To add a control to a window, simply set it as an attribute of the window.

    ::

        from vanilla import Window, Button, TextBox

        class WindowDemo:

            def __init__(self):
                self.w = Window((200, 70), "Window Demo")
                self.w.myButton = Button((10, 10, -10, 20), "My Button")
                self.w.myTextBox = TextBox((10, 40, -10, 17), "My Text Box")
                self.w.open()

        WindowDemo()

    No special naming is required for the attributes. However, each attribute
    must have a unique name.

    **posSize** Tuple of form *(left, top, width, height)* representing the
    position and size of the window. It may also be a tuple of form  *(width, height)*.
    In this case, the window will be positioned on screen automatically.

    **title** The title to be set in the title bar of the window.

    **minSize** Tuple of the form *(width, height)* representing the minimum size
    that the window can be resized to.

    **maxSize** Tuple of the form *(width, height)* representing the maximum size
    that the window can be resized to.

    **textured** Boolean value representing if the window should have a textured
    appearance or not.

    **autosaveName** A string representing a unique name for the window.
    If given, this name will be used to store the window position and size in
    the application preferences.

    **closable** Boolean value representing if the window should have a
    close button in the title bar.

    **miniaturizable** Boolean value representing if the window should have a
    minimize button in the title bar.

    **initiallyVisible** Boolean value representing if the window will be
    initially visible. Default is *True*. If *False*, you can show the window later
    by calling `window.show()`.

    **fullScreenMode** An indication of the full screen mode. These are the options:

    +---------------+---------------------------------------------------------------+
    | *None*        | The window does not allow full screen.                        |
    +---------------+---------------------------------------------------------------+
    | *"primary"*   | Corresponds to NSWindowCollectionBehaviorFullScreenPrimary.   |
    +---------------+---------------------------------------------------------------+
    | *"auxiliary"* | Corresponds to NSWindowCollectionBehaviorFullScreenAuxiliary. |
    +---------------+---------------------------------------------------------------+

    **titleVisible** Boolean value indicating if the window title should be displayed.

    **fullSizeContentView** Boolean value indicating if the content view should be the
    full size of the window, including the area underneath the titlebar and toolbar.

    **screen** A `NSScreen`_ object indicating the screen that the window should be
    drawn to. When None the window will be drawn to the main screen.

    .. _NSScreen: https://developer.apple.com/documentation/appkit/nsscreen?language=objc

    """

    def __new__(cls, *args, **kwargs):
        return cls.alloc().init()

    nsWindowStyleMask = NSTitledWindowMask | NSUnifiedTitleAndToolbarWindowMask
    nsWindowClass = NSWindow
    nsWindowLevel = NSNormalWindowLevel

    def __init__(self, posSize, title="", minSize=None, maxSize=None, textured=False,
                autosaveName=None, closable=True, miniaturizable=True, initiallyVisible=True,
                fullScreenMode=None, titleVisible=True, fullSizeContentView=False, screen=None):
        mask = self.nsWindowStyleMask
        if closable:
            mask = mask | NSClosableWindowMask
        if miniaturizable:
            mask = mask | NSMiniaturizableWindowMask
        if minSize or maxSize:
            mask = mask | NSResizableWindowMask
        if textured:
            mask = mask | NSTexturedBackgroundWindowMask
        if fullSizeContentView and osVersionCurrent >= osVersion10_10:
            mask = mask | NSFullSizeContentViewWindowMask
        # start the window
        ## too magical?
        if len(posSize) == 2:
            l = t = 100
            w, h = posSize
            cascade = True
        else:
            l, t, w, h = posSize
            cascade = False
        if screen is None:
            screen = NSScreen.mainScreen()
        frame = _calcFrame(screen.visibleFrame(), ((l, t), (w, h)))
        self._window = self.nsWindowClass.alloc().initWithContentRect_styleMask_backing_defer_screen_(
            frame, mask, NSBackingStoreBuffered, False, screen)
        if autosaveName is not None:
            # This also sets the window frame if it was previously stored.
            # Make sure we do this before cascading.
            self._window.setFrameAutosaveName_(autosaveName)
        if cascade:
            self._cascade()
        if minSize is not None:
            self._window.setMinSize_(minSize)
        if maxSize is not None:
            self._window.setMaxSize_(maxSize)
        self._window.setTitle_(title)
        self._window.setLevel_(self.nsWindowLevel)
        self._window.setReleasedWhenClosed_(False)
        self._window.setDelegate_(self)
        self._autoLayoutViews = {}
        self._bindings = {}
        self._initiallyVisible = initiallyVisible
        # full screen mode
        if osVersionCurrent >= osVersion10_7:
            if fullScreenMode is None:
                pass
            elif fullScreenMode == "primary":
                self._window.setCollectionBehavior_(NSWindowCollectionBehaviorFullScreenPrimary)
            elif fullScreenMode == "auxiliary":
                self._window.setCollectionBehavior_(NSWindowCollectionBehaviorFullScreenAuxiliary)
        # titlebar visibility
        if osVersionCurrent >= osVersion10_10:
            if not titleVisible:
                self._window.setTitleVisibility_(NSWindowTitleHidden)
            else:
                self._window.setTitleVisibility_(NSWindowTitleVisible)
        # full size content view
        if fullSizeContentView and osVersionCurrent >= osVersion10_10:
            self._window.setTitlebarAppearsTransparent_(True)

    def _testForDeprecatedAttributes(self):
        from warnings import warn
        if hasattr(self, "_nsWindowStyleMask"):
            warn(DeprecationWarning("The _nsWindowStyleMask attribute is deprecated. Use the nsWindowStyleMask attribute."))
            self.nsWindowStyleMask = self._nsWindowStyleMask
        if hasattr(self, "_nsWindowClass"):
            warn(DeprecationWarning("The _nsWindowClass attribute is deprecated. Use the nsWindowClass attribute."))
            self.nsWindowClass = self._nsWindowClass
        if hasattr(self, "_nsWindowLevel"):
            warn(DeprecationWarning("The _nsWindowLevel attribute is deprecated. Use the nsWindowLevel attribute."))
            self.nsWindowLevel = self._nsWindowLevel

    def _cascade(self):
        allLeftTop = []
        for other in NSApp().orderedWindows():
            if other == self._window:
                continue
            (oL, oB), (oW, oH) = other.frame()
            allLeftTop.append((oL, oB + oH))
        (sL, sB), (sW, sH) = self._window.frame()
        leftTop = sL, sB + sH
        while leftTop in allLeftTop:
            leftTop = self._window.cascadeTopLeftFromPoint_(leftTop)
        self._window.setFrameTopLeftPoint_(leftTop)

    def _breakCycles(self):
        _breakCycles(self._window.contentView())
        drawers = self._window.drawers()
        if drawers is not None:
            for drawer in drawers:
                _breakCycles(drawer.contentView())

    def _getContentView(self):
        return self._window.contentView()

    def __setattr__(self, attr, value):
        _setAttr(Window, self, attr, value)

    def __delattr__(self, attr):
        _delAttr(Window, self, attr)

    @python_method
    def assignToDocument(self, document):
        """
        Add this window to the list of windows associated with a document.

        **document** should be a `NSDocument`_ instance.

        .. _NSDocument: https://developer.apple.com/documentation/appkit/nsdocument?language=objc
        """
        document.addWindowController_(self.getNSWindowController())

    def getNSWindow(self):
        """
        Return the `NSWindow`_ that this Vanilla object wraps.

        .. _NSWindow: https://developer.apple.com/documentation/appkit/nswindow?language=objc
        """
        return self._window

    def getNSWindowController(self):
        """
        Return an `NSWindowController`_ for the `NSWindow`_ that this Vanilla
        object wraps, creating a one if needed.

        .. _NSWindowController: https://developer.apple.com/documentation/appkit/nswindowcontroller?language=objc
        """
        controller = self._window.windowController()
        if controller is None:
            controller = NSWindowController.alloc().initWithWindow_(self._window)
        return controller

    def open(self):
        """
        Open the window.
        """
        if self._window is None:
            raise ValueError("can't re-open a window")
        if self._initiallyVisible:
            self.show()
        # We retain ourselves to ensure we don't go away, even if our
        # caller doesn't keep a reference. It's balanced by a release
        # in windowWillClose_().
        self.retain()
        self._validateMinMaxSize()

    def _validateMinMaxSize(self):
        # warn when the min size is bigger then the initial window size
        # or when the max size is smaller then the initial window size
        size = self._window.frame().size
        minSize = self._window.minSize()
        maxSize = self._window.maxSize()
        if size.width < minSize.width or size.height < minSize.height:
            from warnings import warn
            warn("The windows `minSize` is bigger then the initial size.", VanillaWarning)
        elif size.width > maxSize.width or size.height > maxSize.height:
            from warnings import warn
            warn("The windows `maxSize` is bigger then the initial size.", VanillaWarning)

    def close(self):
        """
        Close the window.

        Once a window has been closed it can not be re-opened.
        """
        if self._window.isSheet():
            NSApp().endSheet_(self._window)
            self._window.orderOut_(None)
        self._window.close()

    def hide(self):
        """
        Hide the window.
        """
        self._window.orderOut_(None)

    def show(self):
        """
        Show the window if it is hidden.
        """
        self._window.makeKeyAndOrderFront_(None)

    def makeKey(self):
        """
        Make the window the key window.
        """
        self._window.makeKeyWindow()

    def makeMain(self):
        """
        Make the window the main window.
        """
        self._window.makeMainWindow()

    @python_method
    def setTitle(self, title):
        """
        Set the title in the window's title bar.

        **title** should be a string.
        """
        self._window.setTitle_(title)

    def getTitle(self):
        """
        The title in the window's title bar.
        """
        return self._window.title()

    def select(self):
        """
        Select the window if it is not the currently selected window.
        """
        self._window.makeKeyWindow()

    def isVisible(self):
        """
        A boolean value representing if the window is visible or not.
        """
        return self._window.isVisible()

    def _calculateTitlebarHeight(self):
        # Note: this will include the toolbar height if there is one
        contentFrame = self._window.contentRectForFrameRect_(self._window.frame())
        windowFrame = self._window.frame()
        contentHeight = contentFrame.size[1]
        windowHeight = windowFrame.size[1]
        return windowHeight - contentHeight

    def getPosSize(self):
        """
        A tuple of form *(left, top, width, height)* representing the window's
        position and size.
        """
        frame = self._window.frame()
        l, t, w, h = _flipFrame(self._window.screen().visibleFrame(), frame)
        titlebarHeight = self._calculateTitlebarHeight()
        t += titlebarHeight
        h -= titlebarHeight
        return (l, t, w, h)

    @python_method
    def setPosSize(self, posSize, animate=True):
        """
        Set the position and size of the window.

        **posSize** A tuple of form *(left, top, width, height)*.
        """
        titlebarHeight = self._calculateTitlebarHeight()
        l, t, w, h = posSize
        t -= titlebarHeight
        h += titlebarHeight
        screenFrame = self._window.screen().visibleFrame()
        # if the top is less than zero, force it to zero.
        # otherwise the window will be thrown to the bottom
        # of the screen.
        if t < 0:
            t = 0
            # the screen frame could have a bottom
            # value that is not zero. this will cause
            # an error if (and only if) a window is
            # being positioned at the top of the screen.
            # so, adjust it.
            (sL, sB), (sW, sH) = screenFrame
            screenFrame = ((sL, 0), (sW, sH + sB))
        frame = _calcFrame(screenFrame, ((l, t), (w, h)), absolutePositioning=True)
        self._window.setFrame_display_animate_(frame, True, animate)

    @python_method
    def addAutoPosSizeRules(self, rules, metrics=None):
        """
        Add auto layout rules for controls/view in this view.

        **rules** must by a list of rule definitions.
        Rule definitions may take two forms:

        * strings that follow the `Visual Format Language`_
        * dictionaries with the following key/value pairs:

        +---------------------------+-------------------------------------------------------------+
        | key                       | value                                                       |
        +===========================+=============================================================+
        | *"view1"*                 | The vanilla wrapped view for the left side of the rule.     |
        +---------------------------+-------------------------------------------------------------+
        | *"attribute1"*            | The attribute of the view for the left side of the rule.    |
        |                           | See below for options.                                      |
        +---------------------------+-------------------------------------------------------------+
        | *"relation"* (optional)   | The relationship between the left side of the rule          |
        |                           | and the right side of the rule. See below for options.      |
        |                           | The default value is `"=="`.                                |
        +---------------------------+-------------------------------------------------------------+
        | *"view2"*                 | The vanilla wrapped view for the right side of the rule.    |
        +---------------------------+-------------------------------------------------------------+
        | *"attribute2"*            | The attribute of the view for the right side of the rule.   |
        |                           | See below for options.                                      |
        +---------------------------+-------------------------------------------------------------+
        | *"multiplier"* (optional) | The constant multiplied with the attribute on the right     |
        |                           | side of the rule as part of getting the modified attribute. |
        |                           | The default value is `1`.                                   |
        +---------------------------+-------------------------------------------------------------+
        | *"constant"* (optional)   | The constant added to the multiplied attribute value on     |
        |                           | the right side of the rule to yield the final modified      |
        |                           | attribute. The default value is `0`.                        |
        +---------------------------+-------------------------------------------------------------+

        The `attribute1` and `attribute2` options are:

        +-------------------+--------------------------------+
        | value             | AppKit equivalent              |
        +===================+================================+
        | *"left"*          | NSLayoutAttributeLeft          |
        +-------------------+--------------------------------+
        | *"right"*         | NSLayoutAttributeRight         |
        +-------------------+--------------------------------+
        | *"top"*           | NSLayoutAttributeTop           |
        +-------------------+--------------------------------+
        | *"bottom"*        | NSLayoutAttributeBottom        |
        +-------------------+--------------------------------+
        | *"leading"*       | NSLayoutAttributeLeading       |
        +-------------------+--------------------------------+
        | *"trailing"*      | NSLayoutAttributeTrailing      |
        +-------------------+--------------------------------+
        | *"width"*         | NSLayoutAttributeWidth         |
        +-------------------+--------------------------------+
        | *"height"*        | NSLayoutAttributeHeight        |
        +-------------------+--------------------------------+
        | *"centerX"*       | NSLayoutAttributeCenterX       |
        +-------------------+--------------------------------+
        | *"centerY"*       | NSLayoutAttributeCenterY       |
        +-------------------+--------------------------------+
        | *"baseline"*      | NSLayoutAttributeBaseline      |
        +-------------------+--------------------------------+
        | *"lastBaseline"*  | NSLayoutAttributeLastBaseline  |
        +-------------------+--------------------------------+
        | *"firstBaseline"* | NSLayoutAttributeFirstBaseline |
        +-------------------+--------------------------------+

        Refer to the `NSLayoutAttribute documentation`_ for the information
        about what each of these do.

        The `relation` options are:

        +--------+------------------------------------+
        | value  | AppKit equivalent                  |
        +========+====================================+
        | *"<="* | NSLayoutRelationLessThanOrEqual    |
        +--------+------------------------------------+
        | *"=="* | NSLayoutRelationEqual              |
        +--------+------------------------------------+
        | *">="* | NSLayoutRelationGreaterThanOrEqual |
        +--------+------------------------------------+

        Refer to the `NSLayoutRelation documentation`_ for the information
        about what each of these do.

        **metrics** may be either **None** or a dict containing
        key value pairs representing metrics keywords used in the
        rules defined with strings.

        .. _Visual Format Language: https://developer.apple.com/library/archive/documentation/UserExperience/Conceptual/AutolayoutPG/VisualFormatLanguage.html#//apple_ref/doc/uid/TP40010853-CH27-SW1
        .. _NSLayoutAttribute documentation: https://developer.apple.com/documentation/uikit/nslayoutattribute?language=objc
        .. _NSLayoutRelation documentation: https://developer.apple.com/documentation/uikit/nslayoutrelation?language=objc

        """
        _addAutoLayoutRules(self, rules, metrics)

    def center(self):
        """
        Center the window within the screen.
        """
        self._window.center()

    @python_method
    def move(self, x, y, animate=True):
        """
        Move the window by **x** units and **y** units.
        """
        (l, b), (w, h) = self._window.frame()
        l = l + x
        b = b - y
        self._window.setFrame_display_animate_(((l, b), (w, h)), True, animate)

    @python_method
    def resize(self, width, height, animate=True):
        """
        Change the size of the window to **width** and **height**.
        """
        l, t, w, h = self.getPosSize()
        self.setPosSize((l, t, width, height), animate)

    @python_method
    def setDefaultButton(self, button):
        """
        Set the default button in the window.

        **button** will be bound to the Return and Enter keys.
        """
        if not isinstance(button, VanillaBaseControl):
            raise VanillaError("invalid object")
        cell = button._nsObject.cell()
        self._window.setDefaultButtonCell_(cell)

    @python_method
    def bind(self, event, callback):
        """
        Bind a callback to an event.

        **event** A string representing the desired event. The options are:

        +-------------------+----------------------------------------------------------------------+
        | *"should close"*  | Called when the user attempts to close the window. This must return  |
        |                   | a bool indicating if the window should be closed or not.             |
        +-------------------+----------------------------------------------------------------------+
        | *"close"*         | Called immediately before the window closes.                         |
        +-------------------+----------------------------------------------------------------------+
        | *"move"*          | Called immediately after the window is moved.                        |
        +-------------------+----------------------------------------------------------------------+
        | *"resize"*        | Called immediately after the window is resized.                      |
        +-------------------+----------------------------------------------------------------------+
        | *"became main"*   | Called immediately after the window has become the main window.      |
        +-------------------+----------------------------------------------------------------------+
        | *"resigned main"* | Called immediately after the window has lost its main window status. |
        +-------------------+----------------------------------------------------------------------+
        | *"became key"*    | Called immediately after the window has become the key window.       |
        +-------------------+----------------------------------------------------------------------+
        | *"resigned key"*  | Called immediately after the window has lost its key window status.  |
        +-------------------+----------------------------------------------------------------------+

        *For more information about main and key windows, refer to the Cocoa `documentation`_
        on the subject.*

        .. _documentation: https://developer.apple.com/documentation/Cocoa/Conceptual/WinPanel/Concepts/ChangingMainKeyWindow.html

        **callback** The callback that will be called when the event occurs.
        It should accept a *sender* argument which will be the Window that called the callback.

        ::

            from vanilla import Window

            class WindowBindDemo:

                def __init__(self):
                    self.w = Window((200, 200))
                    self.w.bind("move", self.windowMoved)
                    self.w.open()

                def windowMoved(self, sender):
                    print("window moved!", sender)

            WindowBindDemo()
        """
        if event not in self._bindings:
            self._bindings[event] = []
        self._bindings[event].append(callback)

    @python_method
    def unbind(self, event, callback):
        """
        Unbind a callback from an event.

        **event** A string representing the desired event.
        Refer to *bind* for the options.

        **callback** The callback that has been bound to the event.
        """
        self._bindings[event].remove(callback)

    @python_method
    def _alertBindings(self, key):
        # test to see if the attr exists.
        # this is necessary because NSWindow
        # can move the window (and therefore
        # call the delegate method which calls
        # this method) before the super
        # call in __init__ is complete.
        returnValues = []
        if hasattr(self, "_bindings"):
            if key in self._bindings:
                for callback in self._bindings[key]:
                    value = callback(self)
                    if value is not None:
                        # elimitate None return value
                        returnValues.append(value)
        return all(returnValues)

    def windowWillClose_(self, notification):
        self.hide()
        self._alertBindings("close")
        # remove all bindings to prevent circular refs
        if hasattr(self, "_bindings"):
            del self._bindings
        self._breakCycles()
        # We must make sure that the window does _not_ get deallocated during
        # windowWillClose_, or weird things happen, such as that the window
        # below this window doesn't always properly gets activated. (For reference:
        # this happens when closing with cmd-W, but not when clicking the close
        # control.)
        # Yet we want to get rid of the NSWindow object here, mostly as a flag
        # so we can disallow re-opening windows. So we retain/autorelease the
        # NSWindow, then get rid of our own reference.
        self._window.retain()
        self._window.autorelease()
        self._window = None    # make sure we can't re-open the window
        self.autorelease()     # see self.open()

    def windowDidBecomeKey_(self, notification):
        self._alertBindings("became key")

    def windowDidResignKey_(self, notification):
        self._alertBindings("resigned key")

    def windowDidBecomeMain_(self, notification):
        self._alertBindings("became main")

    def windowDidResignMain_(self, notification):
        self._alertBindings("resigned main")

    def windowDidMove_(self, notification):
        self._alertBindings("move")

    def windowDidResize_(self, notification):
        self._alertBindings("resize")

    def windowDidEnterFullScreen_(self, notification):
        self._alertBindings("enter full screen")

    def windowWillEnterFullScreen_(self, notification):
        self._alertBindings("will enter full screen")

    def windowDidExitFullScreen_(self, notification):
        self._alertBindings("exit full screen")

    def windowWillExitFullScreen_(self, notification):
        self._alertBindings("will exit full screen")

    def windowShouldClose_(self, notification):
        shouldClose = self._alertBindings("should close")
        if shouldClose is None:
            shouldClose = True
        return shouldClose

    # -------
    # Toolbar
    # -------

    # credit where credit is due: much of this was learned
    # from the PyObjC demo: WSTConnectionWindowControllerClass

    @python_method
    def setToolbarStyle(self, toolbarStyle):
        """
        Set a toolbar style for the window.

        **toolbarStyle** A string represetnting the desired toolbar style

        +------------------+
        | "default"        |
        +------------------+
        | "expanded"       |
        +------------------+
        | "preference"     |
        +------------------+
        | "unified"        |
        +------------------+
        | "unifiedCompact" |
        +------------------+
        """
        if osVersionCurrent >= osVersion10_16:
            toolbarStyleMap = dict(
                default=NSWindowToolbarStyleAutomatic,
                expanded=NSWindowToolbarStyleExpanded,
                preference=NSWindowToolbarStylePreference,
                unified=NSWindowToolbarStyleUnified,
                unifiedCompact=NSWindowToolbarStyleUnifiedCompact,
            )
            self._window.setToolbarStyle_(toolbarStyleMap[toolbarStyle])

    @python_method
    def addToolbar(self, toolbarIdentifier, toolbarItems, addStandardItems=True, displayMode="default", sizeStyle="default", toolbarStyle="default"):
        """
        Add a toolbar to the window.

        **toolbarIdentifier** A string representing a unique name for the toolbar.

        **toolbarItems** An ordered list of dictionaries containing the following items:

        +-------------------------------+---------------------------------------------------------------------------+
        | *itemIdentifier*              | A unique string identifier for the item. This is only used internally.    |
        +-------------------------------+---------------------------------------------------------------------------+
        | *label* (optional)            | The text label for the item. Defaults to *None*.                          |
        +-------------------------------+---------------------------------------------------------------------------+
        | *paletteLabel* (optional)     | The text label shown in the customization palette. Defaults to *label*.   |
        +-------------------------------+---------------------------------------------------------------------------+
        | *toolTip* (optional)          | The tool tip for the item. Defaults to *label*.                           |
        +-------------------------------+---------------------------------------------------------------------------+
        | *imagePath* (optional)        | A file path to an image. Defaults to *None*.                              |
        +-------------------------------+---------------------------------------------------------------------------+
        | *imageNamed* (optional)       | The name of an image already loaded as a `NSImage`_ by the application.   |
        |                               | Defaults to *None*.                                                       |
        +-------------------------------+---------------------------------------------------------------------------+
        | *imageObject* (optional)      | A `NSImage`_ object. Defaults to *None*.                                  |
        +-------------------------------+---------------------------------------------------------------------------+
        | *imageTemplate* (optional)    | A boolean representing if the image should converted to a template image. |
        +-------------------------------+---------------------------------------------------------------------------+
        | *selectable* (optional)       | A boolean representing if the item is selectable or not. The default      |
        |                               | value is _False_. For more information on selectable toolbar items, refer |
        |                               | to Apple's documentation.                                                 |
        +-------------------------------+---------------------------------------------------------------------------+
        | *view* (optional)             | A `NSView`_ object to be used instead of an image. Defaults to *None*.    |
        +-------------------------------+---------------------------------------------------------------------------+
        | *visibleByDefault* (optional) | If the item should be visible by default pass True to this argument.      |
        |                               | If the item should be added to the toolbar only through the customization |
        |                               | palette, use a value of _False_. Defaults to _True_.                      |
        +-------------------------------+---------------------------------------------------------------------------+

        .. _NSImage: https://developer.apple.com/documentation/appkit/nsimage?language=objc

        **addStandardItems** A boolean, specifying whether the standard Cocoa toolbar items
        should be added. Defaults to *True*. If you set it to *False*, you must specify any
        standard items manually in *toolbarItems*, by using the constants from the AppKit module:

        +-------------------------------------------+----------------------------------------------------------------+
        | *NSToolbarSeparatorItemIdentifier*        | The Separator item.                                            |
        +-------------------------------------------+----------------------------------------------------------------+
        | *NSToolbarSpaceItemIdentifier*            | The Space item.                                                |
        +-------------------------------------------+----------------------------------------------------------------+
        | *NSToolbarFlexibleSpaceItemIdentifier*    | The Flexible Space item.                                       |
        +-------------------------------------------+----------------------------------------------------------------+
        | *NSToolbarShowColorsItemIdentifier*       | The Colors item. Shows the color panel.                        |
        +-------------------------------------------+----------------------------------------------------------------+
        | *NSToolbarShowFontsItemIdentifier*        | The Fonts item. Shows the font panel.                          |
        +-------------------------------------------+----------------------------------------------------------------+
        | *NSToolbarCustomizeToolbarItemIdentifier* | The Customize item. Shows the customization palette.           |
        +-------------------------------------------+----------------------------------------------------------------+
        | *NSToolbarPrintItemIdentifier*            | The Print item. Refer to Apple's *NSToolbarItem* documentation |
        |                                           | for more information.                                          |
        +-------------------------------------------+----------------------------------------------------------------+

        **displayMode** A string representing the desired display mode for the toolbar.

        +-------------+
        | "default"   |
        +-------------+
        | "iconLabel" |
        +-------------+
        | "icon"      |
        +-------------+
        | "label"     |
        +-------------+

        **sizeStyle** A string representing the desired size for the toolbar

        +-----------+
        | "default" |
        +-----------+
        | "regular" |
        +-----------+
        | "small"   |
        +-----------+

        **toolbarStyle** A string represetnting the desired toolbar style

        +------------------+
        | "default"        |
        +------------------+
        | "expanded"       |
        +------------------+
        | "preference"     |
        +------------------+
        | "unified"        |
        +------------------+
        | "unifiedCompact" |
        +------------------+

        Returns a dictionary containing the created toolbar items, mapped by itemIdentifier.
        """
        STANDARD_TOOLBAR_ITEMS = [
            NSToolbarFlexibleSpaceItemIdentifier,
            NSToolbarSpaceItemIdentifier,
            NSToolbarSeparatorItemIdentifier,
            NSToolbarCustomizeToolbarItemIdentifier,
            NSToolbarPrintItemIdentifier,
            NSToolbarShowFontsItemIdentifier,
            NSToolbarShowColorsItemIdentifier,
        ]
        # create the reference structures
        self._toolbarItems = {}
        self._toolbarDefaultItemIdentifiers = []
        self._toolbarAllowedItemIdentifiers = []
        self._toolbarCallbackWrappers = {}
        self._toolbarSelectableItemIdentifiers = []
        # create the toolbar items
        for itemData in toolbarItems:
            self._createToolbarItem(itemData)
        if addStandardItems:
            for standardItem in STANDARD_TOOLBAR_ITEMS:
                if standardItem not in self._toolbarAllowedItemIdentifiers:
                    self._toolbarAllowedItemIdentifiers.append(standardItem)
        # create the toolbar
        toolbar = NSToolbar.alloc().initWithIdentifier_(toolbarIdentifier)
        toolbar.setDelegate_(self)
        toolbar.setAllowsUserCustomization_(True)
        toolbar.setAutosavesConfiguration_(True)

        displayModeMap = dict(
            default=NSToolbarDisplayModeDefault,
            iconLabel=NSToolbarDisplayModeIconAndLabel,
            icon=NSToolbarDisplayModeIconOnly,
            label=NSToolbarDisplayModeLabelOnly,
        )
        toolbar.setDisplayMode_(displayModeMap[displayMode])

        sizeStyleMap = dict(
            default=NSToolbarSizeModeDefault,
            regular=NSToolbarSizeModeRegular,
            small=NSToolbarSizeModeSmall
        )
        toolbar.setSizeMode_(sizeStyleMap[sizeStyle])

        self.setToolbarStyle(toolbarStyle)
        self._window.setToolbar_(toolbar)
        # Return the dict of toolbar items, so our caller can choose to
        # keep references to them if needed.
        return self._toolbarItems

    def getToolbarItems(self):
        if hasattr(self, "_toolbarItems"):
            return self._toolbarItems
        return {}

    @python_method
    def addToolbarItem(self, itemData, index=None):
        """
        Add a toolbar item to the windows toolbar.

        **itemData** item description with the same format as a toolbarItem description in `addToolbar`

        **index** An integer, specifying the place to insert the toolbar itemIdentifier.
        """
        if not hasattr(self, "_toolbarItems"):
            raise VanillaError("window has not toolbar")
        itemIdentifier = itemData.get("itemIdentifier")
        self._createToolbarItem(itemData)
        if itemData.get("visibleByDefault", True):
            if index is not None:
                self._toolbarDefaultItemIdentifiers.remove(itemIdentifier)
                self._toolbarDefaultItemIdentifiers.insert(index, itemIdentifier)
            index = self._toolbarDefaultItemIdentifiers.index(itemIdentifier)
            self._window.toolbar().insertItemWithItemIdentifier_atIndex_(itemIdentifier, index)

    @python_method
    def removeToolbarItem(self, itemIdentifier):
        """
        Remove a toolbar item by his identifier.

        **itemIdentifier** A unique string identifier for the removed item.
        """
        if not hasattr(self, "_toolbarItems"):
            raise VanillaError("window has not toolbar")
        if itemIdentifier not in self._toolbarItems:
            raise VanillaError("itemIdentifier %r not in toolbar" % itemIdentifier)
        item = self._toolbarItems[itemIdentifier]
        toolbarItems = self._window.toolbar().items()
        if item in toolbarItems:
            ## it can happen a user changed the toolbar manually
            index = toolbarItems.indexOfObject_(item)
            self._window.toolbar().removeItemAtIndex_(index)
        self._toolbarAllowedItemIdentifiers.remove(itemIdentifier)
        self._toolbarDefaultItemIdentifiers.remove(itemIdentifier)
        del self._toolbarItems[itemIdentifier]

    @python_method
    def _createToolbarItem(self, itemData):
        itemIdentifier = itemData.get("itemIdentifier")
        if itemIdentifier is None:
            raise VanillaError("toolbar item data must contain a unique itemIdentifier string")
        if itemIdentifier in self._toolbarItems:
            raise VanillaError("toolbar itemIdentifier is not unique: %r" % itemIdentifier)

        if itemIdentifier not in self._toolbarAllowedItemIdentifiers:
            self._toolbarAllowedItemIdentifiers.append(itemIdentifier)
        if itemData.get("visibleByDefault", True):
            self._toolbarDefaultItemIdentifiers.append(itemIdentifier)

        if itemIdentifier.startswith("NS"):
            # no need to create an actual item for a standard Cocoa toolbar item
            return

        label = itemData.get("label")
        paletteLabel = itemData.get("paletteLabel", label)
        toolTip = itemData.get("toolTip", label)
        imagePath = itemData.get("imagePath")
        imageNamed = itemData.get("imageNamed")
        imageObject = itemData.get("imageObject")
        imageTemplate = itemData.get("imageTemplate")
        view = itemData.get("view")
        callback = itemData.get("callback", None)
        # create the NSImage if needed
        if imagePath is not None:
            image = NSImage.alloc().initWithContentsOfFile_(imagePath)
        elif imageNamed is not None:
            image = NSImage.imageNamed_(imageNamed)
        elif imageObject is not None:
            image = imageObject
        else:
            image = None
        toolbarItem = NSToolbarItem.alloc().initWithItemIdentifier_(itemIdentifier)
        toolbarItem.setLabel_(label)
        toolbarItem.setPaletteLabel_(paletteLabel)
        toolbarItem.setToolTip_(toolTip)
        if image is not None:
            if imageTemplate is not None:
                # only change the image template setting if its either True or False
                image.setTemplate_(imageTemplate)
            toolbarItem.setImage_(image)
        elif view is not None:
            toolbarItem.setView_(view)
            toolbarItem.setMinSize_(view.frame().size)
            toolbarItem.setMaxSize_(view.frame().size)
        if callback is not None:
            target = VanillaCallbackWrapper(callback)
            toolbarItem.setTarget_(target)
            toolbarItem.setAction_("action:")
            self._toolbarCallbackWrappers[itemIdentifier] = target
        if itemData.get("selectable", False):
            self._toolbarSelectableItemIdentifiers.append(itemIdentifier)
        self._toolbarItems[itemIdentifier] = toolbarItem

    # Toolbar delegate methods

    def toolbarDefaultItemIdentifiers_(self, anIdentifier):
        return self._toolbarDefaultItemIdentifiers

    def toolbarAllowedItemIdentifiers_(self, anIdentifier):
        return self._toolbarAllowedItemIdentifiers

    def toolbar_itemForItemIdentifier_willBeInsertedIntoToolbar_(self, toolbar, itemIdentifier, flag):
        return self._toolbarItems.get(itemIdentifier)

    def toolbarSelectableItemIdentifiers_(self, toolbar):
        return self._toolbarSelectableItemIdentifiers


class FloatingWindow(Window):

    """
    A window that floats above all other windows.

    .. image:: /_images/FloatingWindow.png

    To add a control to a window, simply set it as an attribute of the window.

    ::

        from vanilla import FloatingWindow, Button, TextBox

        class FloatingWindowDemo:

            def __init__(self):
                self.w = FloatingWindow((200, 70), "FloatingWindow Demo")
                self.w.myButton = Button((10, 10, -10, 20), "My Button")
                self.w.myTextBox = TextBox((10, 40, -10, 17), "My Text Box")
                self.w.open()

        FloatingWindowDemo()

    No special naming is required for the attributes. However, each attribute
    must have a unique name.

    **posSize** Tuple of form *(left, top, width, height)* representing the position
    and size of the window. It may also be a tuple of form *(width, height)*.
    In this case, the window will be positioned on screen automatically.

    **title** The title to be set in the title bar of the window.

    **minSize** Tuple of the form *(width, height)* representing the minimum size
    that the window can be resized to.

    **maxSize** Tuple of the form *(width, height)* representing the maximum size
    that the window can be resized to.

    **textured** Boolean value representing if the window should have a textured
    appearance or not.

    **autosaveName** A string representing a unique name for the window. If given,
    this name will be used to store the window position and size in the application
    preferences.

    **closable** Boolean value representing if the window should have a close button
    in the title bar.

    **screen** A `NSScreen`_ object indicating the screen that the window
    should be drawn to. When None the window will be drawn to the main screen.

    .. _NSScreen: https://developer.apple.com/documentation/appkit/nsscreen?language=objc
    """

    nsWindowStyleMask = NSTitledWindowMask | NSUtilityWindowMask
    nsWindowClass = NSPanel
    nsWindowLevel = NSFloatingWindowLevel

    def __init__(self, posSize, title="", minSize=None, maxSize=None,
            textured=False, autosaveName=None, closable=True,
            initiallyVisible=True, screen=None):
        super(FloatingWindow, self).__init__(posSize, title, minSize, maxSize,
                textured, autosaveName, closable, initiallyVisible=initiallyVisible, screen=screen)
        self._window.setBecomesKeyOnlyIfNeeded_(True)

    def show(self):
        """
        Show the window if it is hidden.
        """
        # don't make key!
        self._window.orderFront_(None)


class HUDFloatingWindow(FloatingWindow):

    """
    A window that floats above all other windows and has the HUD appearance.

    .. image:: /_images/HUDFloatingWindow.png

    To add a control to a window, simply set it as an attribute of the window.

    ::

        from vanilla import *

        class HUDFloatingWindowDemo:

            def __init__(self):
                self.w = HUDFloatingWindow((200, 70), "HUDFloatingWindow Demo")
                self.w.myButton = Button((10, 10, -10, 20), "My Button")
                self.w.myTextBox = TextBox((10, 40, -10, 17), "My Text Box")
                self.w.open()

        HUDFloatingWindowDemo()

    No special naming is required for the attributes. However, each attribute
    must have a unique name.

    **posSize** Tuple of form *(left, top, width, height)* representing the position
    and size of the window. It may also be a tuple of form *(width, height)*.
    In this case, the window will be positioned on screen automatically.

    **title** The title to be set in the title bar of the window.

    **minSize** Tuple of the form *(width, height)* representing the minimum size
    that the window can be resized to.

    **maxSize** Tuple of the form *(width, height)* representing the maximum size
    that the window can be resized to.

    **textured** Boolean value representing if the window should have a textured
    appearance or not.

    **autosaveName** A string representing a unique name for the window.
    If given, this name will be used to store the window position and size in
    the application preferences.

    **closable** Boolean value representing if the window should have a close button
    in the title bar.

    **screen** A `NSScreen`_ object indicating the screen that the window
    should be drawn to. When None the window will be drawn to the main screen.

    .. _NSScreen: https://developer.apple.com/documentation/appkit/nsscreen?language=objc
    """

    nsWindowStyleMask = NSHUDWindowMask | NSUtilityWindowMask | NSTitledWindowMask | NSBorderlessWindowMask


class Sheet(Window):

    """
    A window that is attached to another window.

    .. image:: /_images/Sheet.png

    To add a control to a sheet, simply set it as an attribute of the sheet.::

        from vanilla import Window, Sheet, Button

        class SheetDemo:

            def __init__(self):
                self.w = Window((240, 140), "Sheet Demo")
                self.w.openSheet = Button((10, -30, -10, 20),
                            "open sheet", callback=self.openSheetCallback)
                self.w.open()

            def openSheetCallback(self, sender):
                self.sheet = Sheet((160, 70), self.w)
                self.sheet.closeSheet = Button((10, -30, -10, 20),
                            "close sheet", callback=self.closeSheetCallback)
                self.sheet.open()

            def closeSheetCallback(self, sender):
                self.sheet.close()
                del self.sheet

        SheetDemo()

    No special naming is required for the attributes. However, each attribute
    must have a unique name.

    **posSize** Tuple of form *(width, height)* representing the size of the sheet.

    **parentWindow** The window that the sheet should be attached to.

    **minSize** Tuple of the form *(width, height)* representing the minimum size
    that the sheet can be resized to.

    **maxSize** Tuple of the form *(width, height)* representing the maximum size
    that the sheet can be resized to.

    **autosaveName** A string representing a unique name for the sheet. If given,
    this name will be used to store the sheet size in the application preferences.
    """

    def __init__(self, posSize, parentWindow, minSize=None, maxSize=None,
            autosaveName=None):
        if isinstance(parentWindow, Window):
            parentWindow = parentWindow._window
        self.parentWindow = parentWindow
        textured = bool(parentWindow.styleMask() & NSTexturedBackgroundWindowMask)
        super(Sheet, self).__init__(posSize, "", minSize, maxSize, textured,
                autosaveName=autosaveName)

    def open(self):
        """
        Open the window.
        """
        parentWindow = self.parentWindow
        NSApp().beginSheet_modalForWindow_modalDelegate_didEndSelector_contextInfo_(
            self._window, parentWindow, None, None, 0)
        # See Window.open():
        self.retain()
        self._validateMinMaxSize()



from __future__ import print_function
import time
import os
import sys
from Foundation import NSObject, NSURL, NSString
from AppKit import NSApplication, NSView, NSColor, NSImage, NSCursor, NSSegmentedControl, NSSegmentSwitchTrackingSelectOne, NSRectFill, NSToolbarPrintItemIdentifier, NSToolbarFlexibleSpaceItemIdentifier, NSToolbarCustomizeToolbarItemIdentifier, NSImageNameInfo
import vanilla
try:
    reload(vanilla)
except NameError:
    # the built-in 'reload' was moved to importlib with Python 3.4
    from importlib import reload
    reload(vanilla)
from vanilla import *
from vanilla.test.testStackView import TestStackView
from vanilla.test.testGridView import TestGridView

import objc
objc.options.verbose = True
vanillaPath = os.path.realpath(vanilla.__file__)
vanillaPath = os.path.dirname(os.path.dirname(os.path.dirname(vanillaPath)))
iconPath = os.path.join(vanillaPath, "Data", "testIcon.tif")
iconName = None
if not os.path.exists(iconPath):
    iconPath = None
    iconName = NSImageNameInfo

sizeStyles = ["regular", "small", "mini"]

listOptions = list(sys.modules.keys())
sortedListOptions = sorted(listOptions)


class BaseTest(object):

    def drawGrid(self):
        w, h = self.w.getPosSize()[2:]
        increment = 10
        for i in range(int(w/increment)):
            if i == 0:
                continue
            attrName = "vline%d" % i
            line = VerticalLine((increment*i, 0, 1, h))
            setattr(self.w, attrName, line)
        for i in range(int(h/increment)):
            if i == 0:
                continue
            attrName = "hline%d" % i
            line = HorizontalLine((0, increment*i, w, 1))
            setattr(self.w, attrName, line)

    def basicCallback(self, sender):
        print(sender)

    def titleCallback(self, sender):
        print(sender, sender.getTitle())

    def getCallback(self, sender):
        print(sender, sender.get())


class WindowTest(BaseTest):

    def __init__(self, textured=False):
        self.textured = textured
        self.w = Window((200, 130), "Window Test", textured=textured)
        self.w.windowButton = Button((10, 10, -10, 20), "Window", callback=self.windowCallback)
        self.w.sheetButton = Button((10, 40, -10, 20), "Sheet", callback=self.sheetCallback)
        self.w.drawerButton = Button((10, 70, -10, 20), "Drawer", callback=self.drawerCallback)
        self.w.floatButton = Button((10, 100, -10, 20), "Floating Window", callback=self.floatCallback)
        self.w.open()

    def windowCallback(self, sender):
        WindowTest(not self.textured)

    def sheetCallback(self, sender):
        self.sheet = Sheet((300, 100), self.w)
        self.sheet.closeButton = Button((10, -30, -10, 20), "Close", callback=self._closeSheet)
        self.sheet.open()

    def _closeSheet(self, sender):
        self.sheet.close()
        del self.sheet

    def drawerCallback(self, sender):
        if not hasattr(self, "drawer1"):
            self.drawer1 = Drawer((50, 50), self.w, preferredEdge="left")
            self.drawer2 = Drawer((50, 50), self.w, preferredEdge="top")
            self.drawer3 = Drawer((50, 50), self.w, preferredEdge="right")
            self.drawer4 = Drawer((50, 50), self.w, preferredEdge="bottom")
        self.drawer1.toggle()
        self.drawer2.toggle()
        self.drawer3.toggle()
        self.drawer4.toggle()

    def floatCallback(self, sender):
        floater = FloatingWindow((100, 100))
        floater.open()


class TextTest(BaseTest):

    def __init__(self, drawGrid=False):
        self.w = Window((440, 190), "Text Test")

        _top = 10
        top = _top

        textSizeStyles = [("regular", 17), ("small", 14), ("mini", 11)]
        for sizeStyle, height in textSizeStyles:
            attrName = "TextBox_%s" % sizeStyle
            button = TextBox((10, top, 100, height), attrName, sizeStyle=sizeStyle)
            setattr(self.w, attrName, button)
            top += 30

        textSizeStyles = [("regular", 22), ("small", 19), ("mini", 15)]
        for sizeStyle, height in textSizeStyles:
            attrName = "SearchBox_%s" % sizeStyle
            button = SearchBox((10, top, 100, height), attrName, callback=self.getCallback, sizeStyle=sizeStyle)
            setattr(self.w, attrName, button)
            top += 30

        top = _top
        textSizeStyles = [("regular", 22), ("small", 19), ("mini", 16)]
        for sizeStyle, height in textSizeStyles:
            attrName = "EditText_%s" % sizeStyle
            button = EditText((120, top, 100, height), attrName, callback=self.getCallback, sizeStyle=sizeStyle)
            setattr(self.w, attrName, button)
            top += 30

        textSizeStyles = [("regular", 21), ("small", 17), ("mini", 14)]
        for sizeStyle, height in textSizeStyles:
            attrName = "ComboBox_%s" % sizeStyle
            button = ComboBox((120, top, 100, height), items=listOptions, callback=self.getCallback, sizeStyle=sizeStyle)
            setattr(self.w, attrName, button)
            top += 30

        self.w.TextEditor = TextEditor((240, 10, 190, 170), sys.copyright, callback=self.getCallback)

        if drawGrid:
            self.drawGrid()

        self.w.open()


class ButtonTest(BaseTest):

    def __init__(self, drawGrid=False):
        self.w = Window((440, 800), "Button Test")

        _top = 10
        top = _top

        buttonSizeStyles = [("regular", 20), ("small", 17), ("mini", 14)]
        for sizeStyle, height in buttonSizeStyles:
            attrName = "Button_%s" % sizeStyle
            button = Button((10, top, 150, height), attrName, callback=self.titleCallback, sizeStyle=sizeStyle)
            setattr(self.w, attrName, button)
            top += 30

        height = 20
        for sizeStyle in sizeStyles:
            attrName = "SquareButton_%s" % sizeStyle
            button = SquareButton((10, top, 150, height), attrName, callback=self.titleCallback, sizeStyle=sizeStyle)
            setattr(self.w, attrName, button)
            top += 30

        settings = [(None, False, "top"),
                    (None, True, "top"),
                    ("bop", True, "top"),
                    ("bop", True, "bottom"),
                    ("bop", True, "left"),
                    ("bop", True, "right"),]
        for title, bordered, imagePosition in settings:
            attrName = "ImageButton_%s_%s_%s" % (title, bordered, imagePosition)
            button = ImageButton((10, top, 150, 50), title=title, imagePath=iconPath, imageNamed=iconName, bordered=bordered, imagePosition=imagePosition, callback=self.basicCallback)
            setattr(self.w, attrName, button)
            top += 60

        segmentedControlSizeStyles = [("regular", 20), ("small", 17), ("mini", 14)]
        descriptions = [{"title":"One"}, {"title":"Two"}, {"title":"3", "enabled":False}]
        for sizeStyle, height in segmentedControlSizeStyles:
            attrName = "SegmentedButton_%s" % sizeStyle
            button = SegmentedButton((10, top, 150, height), descriptions, callback=None, sizeStyle=sizeStyle)
            setattr(self.w, attrName, button)
            top += 30

        top = _top
        _left = 170
        left = _left
        height = 100

        sliderSizeStyles = [("regular", 15), ("small", 11), ("mini", 10)]
        for sizeStyle, width in sliderSizeStyles:
            attrName = "VSlider_noTicks_%s" % sizeStyle
            button = Slider((left, top, width, height), 0, 100, callback=self.getCallback, sizeStyle=sizeStyle)
            setattr(self.w, attrName, button)
            left += 30
        sliderSizeStyles = [("regular", 23), ("small", 17), ("mini", 16)]
        for sizeStyle, width in sliderSizeStyles:
            attrName = "VSlider_rightTicks_%s" % sizeStyle
            button = Slider((left, top, width, height), 0, 100, tickMarkCount=10, callback=self.getCallback, sizeStyle=sizeStyle)
            setattr(self.w, attrName, button)
            left += 30
        for sizeStyle, width in sliderSizeStyles:
            attrName = "VSlider_leftTicks_%s" % sizeStyle
            button = Slider((left, top, width, height), 0, 100, tickMarkCount=10, callback=self.getCallback, sizeStyle=sizeStyle)
            button.setTickMarkPosition("left")
            setattr(self.w, attrName, button)
            left += 30

        left = _left
        width = 260
        top = 130
        sliderSizeStyles = [("regular", 15), ("small", 12), ("mini", 10)]
        for sizeStyle, height in sliderSizeStyles:
            attrName = "HSlider_noTicks_%s" % sizeStyle
            button = Slider((left, top, width, height), 0, 100, callback=self.getCallback, sizeStyle=sizeStyle)
            setattr(self.w, attrName, button)
            top += 30
        sliderSizeStyles = [("regular", 24), ("small", 17), ("mini", 16)]
        for sizeStyle, height in sliderSizeStyles:
            attrName = "HSlider_belowTicks_%s" % sizeStyle
            button = Slider((left, top, width, height), 0, 100, tickMarkCount=30, callback=self.getCallback, sizeStyle=sizeStyle)
            setattr(self.w, attrName, button)
            top += 30
        for sizeStyle, height in sliderSizeStyles:
            attrName = "HSlider_aboveTicks_%s" % sizeStyle
            button = Slider((left, top, width, height), 0, 100, tickMarkCount=30, callback=self.getCallback, sizeStyle=sizeStyle)
            button.setTickMarkPosition("top")
            setattr(self.w, attrName, button)
            top += 30

        _top = top

        popupSizeStyles = [("regular", 20), ("small", 17), ("mini", 15)]
        width = 120
        for sizeStyle, height in popupSizeStyles:
            attrName = "PopUpButton_%s" % sizeStyle
            button = PopUpButton((left, top, width, height), listOptions, callback=self.getCallback, sizeStyle=sizeStyle)
            setattr(self.w, attrName, button)
            top += 30

        top = _top
        checkboxSizeStyles = [("regular", 22), ("small", 18), ("mini", 11)]
        width = 125
        left = 300
        for sizeStyle, height in popupSizeStyles:
            attrName = "CheckBox_%s" % sizeStyle
            button = CheckBox((left, top, width, height), attrName, callback=self.getCallback, sizeStyle=sizeStyle)
            setattr(self.w, attrName, button)
            top += 30

        _top = top

        left = _left
        width = 120
        self.w.DLevelIndicator = LevelIndicator((left, top, width, 18), style="discrete",
                    value=5, warningValue=7, criticalValue=9,
                    callback=self.getCallback)
        top += 30
        self.w.DLevelIndicator_ticksAbove = LevelIndicator((left, top, width, 25), style="discrete",
                    value=5, warningValue=7, criticalValue=9,
                    tickMarkPosition="above", minorTickMarkCount=5, majorTickMarkCount=3, callback=self.getCallback)
        top += 30
        self.w.DLevelIndicator_ticksBelow = LevelIndicator((left, top, width, 25), style="discrete",
                    value=5, warningValue=7, criticalValue=9,
                    tickMarkPosition="below", minorTickMarkCount=5, majorTickMarkCount=3, callback=self.getCallback)

        left = 300
        top = _top
        width = 120
        self.w.CLevelIndicator = LevelIndicator((left, top, width, 16), style="continuous",
                    value=5, warningValue=7, criticalValue=9,
                    callback=self.getCallback)
        top += 30
        self.w.CLevelIndicator_ticksAbove = LevelIndicator((left, top, width, 23), style="continuous",
                    value=5, warningValue=7, criticalValue=9,
                    tickMarkPosition="above", minorTickMarkCount=5, majorTickMarkCount=3, callback=self.getCallback)
        top += 30
        self.w.CLevelIndicator_ticksBelow = LevelIndicator((left, top, width, 23), style="continuous",
                    value=5, warningValue=7, criticalValue=9,
                    tickMarkPosition="below", minorTickMarkCount=5, majorTickMarkCount=3, callback=self.getCallback)

        _top = 660
        top = _top
        pathControlSizeStyles = [("regular", 22), ("small", 20), ("mini", 18)]
        left = 10
        width = -10
        urlPath = __file__
        for sizeStyle, height in pathControlSizeStyles:
            attrName = "PathControl_%s" % sizeStyle
            button = PathControl((left, top, width, height), urlPath, callback=None, sizeStyle=sizeStyle)
            setattr(self.w, attrName, button)
            top += 30

        if drawGrid:
            self.drawGrid()

        self.w.open()


class ListTest(BaseTest):

    def __init__(self, drawGrid=False):
        self.w = Window((600, 500), "List Test", minSize=(400, 400))

        simpleList = List((0, 0, 0, 0), listOptions, enableTypingSensitivity=True)

        multiItems = [
            {"name": name, "path": os.path.basename(getattr(module, "__file__", "Unknown"))}
            for name, module in sys.modules.items()
        ]
        columnDescriptions = [
            {"title": "Module Name", "key": "name"},
            {"title": "File Name", "key": "path"}
        ]
        multiList = List((0, 0, 0, 0), multiItems, columnDescriptions=columnDescriptions, enableTypingSensitivity=True)

        if iconPath:
            image = NSImage.alloc().initWithContentsOfFile_(iconPath)
        else:
            image = NSImage.imageNamed_(iconName)

        miscItems = [
            {"slider": 50, "sliderWithTicks" : 50, "checkBox": False, "image" : image, "segment" : 0},
            {"slider": 20, "sliderWithTicks" : 20, "checkBox": True,  "image" : image, "segment" : 1},
            {"slider": 70, "sliderWithTicks" : 70, "checkBox": False, "image" : image, "segment" : 2},
            {"slider": 20, "sliderWithTicks" : 20, "checkBox": True,  "image" : image, "segment" : 0},
            {"slider": 10, "sliderWithTicks" : 10, "checkBox": True,  "image" : image, "segment" : 1},
            {"slider": 90, "sliderWithTicks" : 90, "checkBox": False, "image" : image, "segment" : 2},
        ]
        columnDescriptions = [
            {"title": "SliderListCell", "key": "slider", "cell": SliderListCell()},
            {"title": "SliderListCell", "key": "sliderWithTicks", "cell": SliderListCell(tickMarkCount=10, stopOnTickMarks=True)},
            {"title": "CheckBoxListCell", "key": "checkBox", "cell": CheckBoxListCell()},
            {"title": "ImageListCell", "key": "image", "cell": ImageListCell()},
            {"title": "SegmentedButtonListCell", "key": "segment", "cell": SegmentedButtonListCell([dict(title="0"), dict(title="1"), dict(title="2")]), "binding": "selectedIndex"},
        ]
        miscCellList = List((0, 0, 0, 0), items=miscItems, columnDescriptions=columnDescriptions)

        paneDescriptions = [
            dict(view=simpleList, identifier="simpleList"),
            dict(view=multiList, identifier="multiList"),
            dict(view=miscCellList, identifier="miscCellList"),
        ]

        # only add the ListIndicator tests if the controls are available
        try:
            listIndicatorItems = [
                {"discrete": 3, "continuous": 4, "rating": 1, "relevancy": 9},
                {"discrete": 8, "continuous": 3, "rating": 5, "relevancy": 5},
                {"discrete": 3, "continuous": 7, "rating": 3, "relevancy": 4},
                {"discrete": 2, "continuous": 5, "rating": 4, "relevancy": 7},
                {"discrete": 6, "continuous": 9, "rating": 3, "relevancy": 2},
                {"discrete": 4, "continuous": 0, "rating": 6, "relevancy": 8},
            ]
            columnDescriptions = [
                {"title": "discrete",
                "cell": LevelIndicatorListCell(style="discrete", warningValue=7, criticalValue=9)},
                {"title": "continuous",
                "cell": LevelIndicatorListCell(style="continuous", warningValue=7, criticalValue=9)},
                {"title": "rating",
                "cell": LevelIndicatorListCell(style="rating", maxValue=6)},
                {"title": "relevancy",
                "cell": LevelIndicatorListCell(style="relevancy")},
            ]
            levelIndicatorList = List((0, 0, 0, 0), items=listIndicatorItems, columnDescriptions=columnDescriptions)
            paneDescriptions.append(dict(view=levelIndicatorList, identifier="levelIndicatorList"))
        except NameError:
            pass

        self.w.splitView = SplitView((0, 0, -0, -0), paneDescriptions, isVertical=False)

        if drawGrid:
            self.drawGrid()

        self.w.open()

class BrowserTest(BaseTest):

    def __init__(self, drawGrid=False):
        self.w = Window((440, 500), "Browser Test", minSize=(400, 400))

        import vanilla
        self.w.browser = ObjectBrowser((0, 0, 0, 0), vanilla)

        if drawGrid:
            self.drawGrid()

        self.w.open()


class TestCustomNSView(NSView):

    def viewDidEndLiveResize(self):
        self._recalcSize()

    def _recalcSize(self):
        # XXX Note that this is specific for embedding in a ScrollView,
        # it may behave strangely when used in another context.
        w, h = self.superview().visibleRect()[1]
        self.setFrame_(((0, 0), (w, h)))

    def drawRect_(self, rect):
        if self.inLiveResize():
            self._recalcSize()
        from random import random
        NSColor.redColor().set()
        NSRectFill(self.bounds())
        width, height = self.frame()[1]
        w = width / 5
        h = height / 5
        for xI in range(5):
            for yI in range(5):
                x = xI * w
                y = height - (yI * h) - h
                r = ((x, y), (w, h))
                NSColor.colorWithDeviceRed_green_blue_alpha_(random(), random(), random(), 1.0).set()
                NSRectFill(r)


class ViewTest(BaseTest):

    def __init__(self, drawGrid=False):
        self.w = Window((450, 350), "View Test", minSize=(350, 300))

        self.w.tabs = Tabs((10, 10, 220, 120), ["Small", "Mini"])
        self.w.tabs[0].tabs = Tabs((10, 10, -10, -10), ["One", "Two", "Three"], sizeStyle="small")
        self.w.tabs[1].tabs = Tabs((10, 10, -10, -10), ["One", "Two", "Three"], sizeStyle="mini")

        self.w.box = Box((10, 140, 220, 70), "Box")
        self.w.box.box = Box((10, 10, -10, -10))

        self.scrollViewNSView = TestCustomNSView.alloc().initWithFrame_(((0, 0), (500, 500)))
        self.w.scrollView = ScrollView((240, 10, 200, 200), self.scrollViewNSView, backgroundColor=NSColor.redColor())

        self.splitViewNSView1 = TestCustomNSView.alloc().initWithFrame_(((0, 0), (0, 0)))
        self.splitViewNSView2 = TestCustomNSView.alloc().initWithFrame_(((0, 0), (0, 0)))
        view1 = ScrollView((0, 0, 0, 50), self.splitViewNSView1, autohidesScrollers=True, backgroundColor=NSColor.redColor())
        view2 = ScrollView((0, 0, 0, -10), self.splitViewNSView2, autohidesScrollers=True, backgroundColor=NSColor.redColor())
        paneDescriptions = [
            dict(view=view1, identifier="view1"),
            dict(view=view2, identifier="view2"),
        ]
        self.w.splitView = SplitView((10, 220, -10, -10), paneDescriptions)

        if drawGrid:
            self.drawGrid()

        self.w.open()


class ToolbarTest(BaseTest):

    def __init__(self, drawGrid=False):
        self.w = Window((350, 20), "Toolbar Test", minSize=(250, 20))

        customView = NSSegmentedControl.alloc().initWithFrame_(((0, 0), (100, 30)))
        cell = customView.cell()
        cell.setTrackingMode_(NSSegmentSwitchTrackingSelectOne)
        cell.setSegmentCount_(2)
        cell.setImage_forSegment_(NSCursor.arrowCursor().image(), 0)
        cell.setImage_forSegment_(NSCursor.crosshairCursor().image(), 1)
        customView.sizeToFit()

        toolbarItems = [
            {"itemIdentifier": "Test Item One",
             "label": "Test One",
             "imagePath": iconPath,
             "imageNamed": iconName,
             "callback": self.basicCallback},
            {"itemIdentifier": "Test Item Two",
             "label": "Test Two",
             "imagePath": iconPath,
             "imageNamed": iconName,
             "callback": self.basicCallback},
            {"itemIdentifier": "Test Item Three",
             "imagePath": iconPath,
             "imageNamed": iconName,
             "callback": self.basicCallback},
            {"itemIdentifier": "Test Item Four",
             "label": "Test Four",
             "view": customView,
             "callback": self.basicCallback},
            {"itemIdentifier": NSToolbarPrintItemIdentifier, "visibleByDefault": False},
            {"itemIdentifier": NSToolbarFlexibleSpaceItemIdentifier},
            {"itemIdentifier": NSToolbarCustomizeToolbarItemIdentifier},
        ]

        self.w.addToolbar("Vanilla Test Toolbar", toolbarItems=toolbarItems)

        self.w.open()


class MiscTest(BaseTest):

    def __init__(self, drawGrid=False):
        self.w = Window((150, 180), "Misc. Test")

        self.w.spinner1 = ProgressSpinner((10, 10, 32, 32), sizeStyle="regular")
        self.w.spinner2 = ProgressSpinner((50, 10, 16, 16), sizeStyle="small")

        self.w.bar1 = ProgressBar((10, 50, -10, 16))
        self.w.bar2 = ProgressBar((10, 70, -10, 10), isIndeterminate=True, sizeStyle="small")

        self.w.progressStartButton = Button((10, 90, -10, 20), "Start Progress", callback=self.startProgress)

        self.w.colorWell = ColorWell((10, 130, -10, -10), callback=self.getCallback, color=NSColor.redColor())

        if drawGrid:
            self.drawGrid()

        self.w.open()

    def startProgress(self, sender):
        self.w.spinner1.start()
        self.w.spinner2.start()
        self.w.bar2.start()
        for i in range(10):
            self.w.bar1.increment(10)
            time.sleep(.1)
        time.sleep(.5)
        self.w.spinner1.stop()
        self.w.spinner2.stop()
        self.w.bar2.stop()
        self.w.bar1.set(0)


class _VanillaTestViewForSplitView(NSView):

    def drawRect_(self, rect):
        from AppKit import NSRectFill, NSBezierPath, NSColor
        self.color.set()
        NSRectFill(self.bounds())
        NSColor.blackColor().set()
        p = NSBezierPath.bezierPathWithRect_(self.bounds())
        p.setLineWidth_(10)
        p.stroke()


class TestSplitSubview(VanillaBaseObject):

    def __init__(self, posSize, color):
        self._setupView(_VanillaTestViewForSplitView, posSize)
        self._nsObject.color = color

class TestSplitView(BaseTest):

    def __init__(self, drawGrid=False):
        self.w = Window((600, 500), "", minSize=(300, 250))

        grp = Group((0, 0, 0, 0))
        grp.button = Button((10, 10, -10, 20), "Toggle", self.buttonCallback)

        self.view1 = TestSplitSubview((0, 0, 0, 0), NSColor.redColor())
        paneDescriptions2 = [
            dict(view=self.view1, canCollapse=True, size=50, identifier="pane1"),
            dict(view=grp, identifier="pane2"),
            dict(view=TestSplitSubview((0, 0, 0, 0), NSColor.greenColor()), minSize=50, identifier="pane3"),
            dict(view=TestSplitSubview((0, 0, 0, 0), NSColor.yellowColor()), identifier="pane4"),
        ]
        self.nestedSplit = SplitView((0, 0, 0, 0), paneDescriptions2, isVertical=True)
        paneDescriptions1 = [
            dict(view=self.nestedSplit, identifier="pane5"),
            dict(view=TestSplitSubview((0, 0, 0, 0), NSColor.magentaColor()), minSize=100, size=100, canCollapse=True, identifier="pane6"),
        ]
        self.w.splitView = SplitView((10, 10, -10, -10), paneDescriptions1, isVertical=False)

        if drawGrid:
            self.drawGrid()
        self.w.open()

    def buttonCallback(self, sender):
        self.nestedSplit.togglePane("pane1")


class Test(object):

    def __init__(self):
        self.w = FloatingWindow((200, 300, 120, 400))
        self.w.drawGrid = CheckBox((10, 10, -10, 22), "Draw Grid", value=False)
        self.w.windows = Button((10, 40, -10, 20), "Windows", callback=self.openTestCallback)
        self.w.geometry = Button((10, 70, -10, 20), "Geometry", callback=self.openTestCallback)
        self.w.text = Button((10, 100, -10, 20), "Text", callback=self.openTestCallback)
        self.w.buttons = Button((10, 130, -10, 20), "Buttons", callback=self.openTestCallback)
        self.w.list = Button((10, 160, -10, 20), "List", callback=self.openTestCallback)
        self.w.browser = Button((10, 190, -10, 20), "Browser", callback=self.openTestCallback)
        self.w.view = Button((10, 220, -10, 20), "Views", callback=self.openTestCallback)
        self.w.toolbar = Button((10, 250, -10, 20), "Toolbar", callback=self.openTestCallback)
        self.w.misc = Button((10, 280, -10, 20), "Misc.", callback=self.openTestCallback)
        self.w.split = Button((10, 310, -10, 20), "SplitView", callback=self.openTestCallback)
        self.w.stack = Button((10, 340, -10, 20), "StackView", callback=self.openTestCallback)
        self.w.grid = Button((10, 370, -10, 20), "GridView", callback=self.openTestCallback)
        self.w.open()

    def openTestCallback(self, sender):
        title = sender.getTitle()
        try:
            if title == "Windows":
                WindowTest()
            elif title == "Geometry":
                from vanilla.test.testGeometry import TestGeometry
                TestGeometry()
            elif title == "Text":
                TextTest(self.w.drawGrid.get())
            elif title == "Buttons":
                ButtonTest(self.w.drawGrid.get())
            elif title == "List":
                ListTest(self.w.drawGrid.get())
            elif title == "Browser":
                BrowserTest(self.w.drawGrid.get())
            elif title == "Views":
                ViewTest(self.w.drawGrid.get())
            elif title == "Toolbar":
                ToolbarTest(self.w.drawGrid.get())
            elif title == "Misc.":
                MiscTest(self.w.drawGrid.get())
            elif title == "SplitView":
                TestSplitView(self.w.drawGrid.get())
            elif title == "StackView":
                TestStackView()
            elif title == "GridView":
                TestGridView()
        except:
            import traceback
            print(traceback.format_exc())

if __name__ == "__main__":
    from vanilla.test.testTools import executeVanillaTest
    executeVanillaTest(Test)



import objc
from AppKit import NSView
from vanilla.vanillaBase import VanillaBaseObject, osVersionCurrent, osVersion10_10
from vanilla.dragAndDrop import DropTargetProtocolMixIn

try:
    NSVisualEffectMaterialAppearanceBased
    NSVisualEffectMaterialLight
    NSVisualEffectMaterialDark
    NSVisualEffectMaterialTitlebar
    NSVisualEffectBlendingModeBehindWindow
    NSVisualEffectBlendingModeWithinWindow
except NameError:
    NSVisualEffectMaterialAppearanceBased = 0
    NSVisualEffectMaterialLight = 1
    NSVisualEffectMaterialDark = 2
    NSVisualEffectMaterialTitlebar = 3
    NSVisualEffectBlendingModeBehindWindow = 0
    NSVisualEffectBlendingModeWithinWindow = 1

if osVersionCurrent >= osVersion10_10:
    from AppKit import NSVisualEffectView, CALayer
    _blendingModeMap = {
        "behindWindow" : NSVisualEffectBlendingModeBehindWindow,
        "withinWindow" : NSVisualEffectBlendingModeWithinWindow
    }
else:
    NSVisualEffectView = None


class VanillaGroupViewSubclass(NSView):

    def draggingEntered_(self, draggingInfo):
        return self.vanillaWrapper()._dropCandidateEntered(draggingInfo)

    def draggingUpdated_(self, draggingInfo):
        return self.vanillaWrapper()._dropCandidateUpdated(draggingInfo)

    @objc.signature(b"Z@:@") # PyObjC bug? <- Found in the FontGoogles source.
    def draggingEnded_(self, draggingInfo):
        return self.vanillaWrapper()._dropCandidateEnded(draggingInfo)

    def draggingExited_(self, draggingInfo):
        return self.vanillaWrapper()._dropCandidateExited(draggingInfo)

    def updateDraggingItemsForDrag_(self, draggingInfo):
        return self.vanillaWrapper()._updateDropCandidateImages(draggingInfo)

    def prepareForDragOperation_(self, draggingInfo):
        return self.vanillaWrapper()._prepareForDrop(draggingInfo)

    def performDragOperation_(self, draggingInfo):
        return self.vanillaWrapper()._performDrop(draggingInfo)

    def concludeDragOperation_(self, draggingInfo):
        return self.vanillaWrapper()._finishDrop(draggingInfo)


class Group(VanillaBaseObject, DropTargetProtocolMixIn):

    """
    An invisible container for controls.

    .. image:: /_images/Group.png

    To add a control to a group, simply set it as an attribute of the group.

    ::

        from vanilla import Window, Group, TextBox

        class GroupDemo:

            def __init__(self):
                self.w = Window((150, 50))
                self.w.group = Group((10, 10, -10, -10))
                self.w.group.text = TextBox((0, 0, -0, -0),
                                        "This is a group")
                self.w.open()

        GroupDemo()

    No special naming is required for the attributes. However, each attribute
    must have a unique name.

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"*
    representing the position and size of the group.

    **blendingMode** The blending mode for the window.
    These are the possible options:

    +----------------+-------------------------------------------+
    | None           | No special blending.                      |
    +----------------+-------------------------------------------+
    | "behindWindow" | Blend with the content behind the window. |
    +----------------+-------------------------------------------+
    | "withinWindow" | Blend with the content within the window. |
    +----------------+-------------------------------------------+

    **dropSettings** A drop settings dictionary.
    """

    nsViewClass = VanillaGroupViewSubclass
    nsVisualEffectViewClass = NSVisualEffectView

    def __init__(self, posSize, blendingMode=None, dropSettings=None):
        self._setupView(self.nsViewClass, posSize)
        if blendingMode is not None and osVersionCurrent >= osVersion10_10:
            # the visual effect view is defined as a subview
            # of the basic NSView. this is necessary because
            # within window blending requires a CA layer and
            # we may as well do it for behind window blending
            # as well to make the code in here clear.
            self._visualEffectGroup = _VisualEffectGroup(self.nsVisualEffectViewClass)
            if blendingMode == "withinWindow":
                self._visualAppearanceLayer = CALayer.layer()
                self._nsObject.setLayer_(self._visualAppearanceLayer)
                self._nsObject.setWantsLayer_(True)
            blendingMode = _blendingModeMap[blendingMode]
            self._visualEffectGroup.getNSVisualEffectView().setBlendingMode_(blendingMode)
        if dropSettings is not None:
            self.setDropSettings(dropSettings)

    def __setattr__(self, attr, value):
        # __init__
        if not hasattr(self, "_visualEffectGroup") or attr == "_visualEffectGroup":
            super(Group, self).__setattr__(attr, value)
        # adding a vanilla subview: add to the visual effect view
        elif hasattr(self, "_visualEffectGroup") and isinstance(value, VanillaBaseObject) and hasattr(value, "_posSize"):
            setattr(self._visualEffectGroup, attr, value)
        # fallback: add to the main view
        else:
            super(Group, self).__setattr__(attr, value)

    def __delattr__(self, attr):
        if hasattr(self, attr):
            super(Group, self).__delattr__(attr)
        elif hasattr(self, "_visualEffectView"):
            delattr(self._visualEffectView, attr)

    def getNSView(self):
        """
        Return the `NSView`_ that this object wraps.

        .. _NSView: https://developer.apple.com/documentation/appkit/nsview?language=objc
        """
        return self._nsObject

    def getNSVisualEffectView(self):
        """
        Return the `NSVisualEffectView`_ that this object wraps.

        .. _NSVisualEffectView: https://developer.apple.com/documentation/appkit/nsvisualeffectview?language=objc
        """
        return self._visualEffectGroup.getNSVisualEffectView()


class _VisualEffectGroup(VanillaBaseObject):

    def __init__(self, cls):
        self._setupView(cls, (0, 0, 0, 0))

    def getNSVisualEffectView(self):
        return self._nsObject



import vanilla

def makeItems():
    items = []
    letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    for i, letter in enumerate(letters):
        number = i + 1
        item = dict(
            letter=letter,
            number=number
        )
        items.append(item)
    return items


class Test:

    def __init__(self):
        self.w = vanilla.Window((500, 1))
        self.w.line = vanilla.VerticalLine("auto")

        columnDescriptions = [
            dict(
                identifier="letter",
                title="Letter",
                sortable=True
            ),
            dict(
                identifier="number",
                title="Number",
                sortable=True
            )
        ]
        self.w.text = vanilla.TextBox(
            "auto",
            "Select items on the left and the same items "
            "should be selected on the right regardless "
            "of different sort orders."
        )
        self.w.itemList1 = vanilla.List2(
            "auto",
            items=makeItems(),
            columnDescriptions=columnDescriptions,
            selectionCallback=self.itemList1SelectionCallback
        )
        self.w.itemList2 = vanilla.List2(
            "auto",
            items=makeItems(),
            columnDescriptions=columnDescriptions
        )
        rules = [
            "H:|-[text]-|",
            "H:|-[itemList1]-[itemList2(==itemList1)]-|",
            "V:|-[text]",
            "V:[text]-[itemList1(==300)]-|",
            "V:[text]-[itemList2(==itemList1)]-|",
        ]
        self.w.addAutoPosSizeRules(rules)
        self.w.open()

    def itemList1SelectionCallback(self, sender):
        indexes = self.w.itemList1.getSelectedIndexes()
        self.w.itemList2.setSelectedIndexes(indexes)


if __name__ == "__main__":
    from vanilla.test.testTools import executeVanillaTest
    executeVanillaTest(Test)



import AppKit
import vanilla

class TestGridView:

    def __init__(self):
        self.w = vanilla.Window((0, 0))

        columnDescriptions = [
            dict(
                columnPlacement="trailing"
            ),
            dict(
                width=300,
                columnPlacement="fill"
            )
        ]

        self.editText = vanilla.EditText(
            "auto",
            "Type to test callback.",
            callback=self.editTextCallback
        )
        rows = [
            (
                vanilla.TextBox("auto", "TextBox:"),
                vanilla.TextBox("auto", "Hello")
            ),
            (
                vanilla.TextBox("auto", "EditText:"),
                self.editText
            ),
            (
                vanilla.TextBox("auto", "Button:"),
                vanilla.Button("auto", "Button")
            ),
            (
                vanilla.TextBox("auto", "PopUpButton:"),
                vanilla.PopUpButton("auto", ["PopUpButton"])
            ),
            (
                vanilla.TextBox("auto", "ComboBox:"),
                vanilla.ComboBox("auto", ["ComboBox"])
            ),
            (
                vanilla.TextBox("auto", "CheckBox:"),
                vanilla.CheckBox("auto", "CheckBox 1")
            ),
            dict(
                rowPadding=(0, 0),
                cells=(
                    None,
                    vanilla.CheckBox("auto", "CheckBox 2")
                )
            ),
            dict(
                rowPadding=(0, 0),
                cells=(
                    None,
                    vanilla.CheckBox("auto", "CheckBox 3")
                )
            ),
            (
                vanilla.TextBox("auto", "SegmentedButton:"),
                vanilla.SegmentedButton(
                    "auto",
                    [
                        dict(title="One"),
                        dict(title="Two"),
                        dict(title="Three")
                    ],
                    sizeStyle="regular"
                )
            ),
            (
                vanilla.TextBox("auto", "Slider:"),
                vanilla.Slider("auto", minValue=0, maxValue=100, value=50)
            ),
            dict(
                height=100,
                cells=(
                    vanilla.TextBox("auto", "RadioGroup:"),
                    dict(
                        rowAlignment="none",
                        view=vanilla.RadioGroup("auto", ["One", "Two", "Three"])
                    )
                )
            ),
            dict(
                height=100,
                cells=(
                    vanilla.TextBox("auto", "ColorWell:"),
                    dict(
                        width=50,
                        height=75,
                        view=vanilla.ColorWell("auto", color=AppKit.NSColor.redColor())
                    )
                )
            ),
            dict(
                height=100,
                cells=(
                    vanilla.TextBox("auto", "List:"),
                    dict(
                        columnPlacement="fill",
                        rowPlacement="fill",
                        view=vanilla.List("auto", ["One", "Two", "Three"])
                    )
                )
            ),
            dict(
                height=100,
                cells=(
                    vanilla.TextBox("auto", "TextEditor:"),
                    vanilla.TextEditor("auto", "TextEditor")
                )
            ),
            dict(
                height=100,
                cells=(
                    vanilla.TextBox("auto", "Box:"),
                    vanilla.Box("auto", fillColor=AppKit.NSColor.blueColor(), cornerRadius=20)
                )
            ),
        ]
        self.w.gridView = vanilla.GridView(
            "auto",
            rows,
            columnDescriptions=columnDescriptions,
            columnWidth=150,
            columnSpacing=10,
            rowHeight=25,
            rowPadding=(15, 0)
        )
        self.columnCount = len(columnDescriptions)
        self.rowCount = len(rows)

        # append, insert, etc. buttons

        self.appendColumnButton = vanilla.Button("auto", "Append", callback=self.appendColumnButtonCallback)
        self.insertColumnButton = vanilla.Button("auto", "Insert", callback=self.insertColumnButtonCallback)
        self.removeColumnButton = vanilla.Button("auto", "Remove", callback=self.removeColumnButtonCallback)
        self.moveColumnButton = vanilla.Button("auto", "Move", callback=self.moveColumnButtonCallback)
        self.showColumnButton = vanilla.Button("auto", "Show/Hide", callback=self.showColumnButtonCallback)

        self.appendRowButton = vanilla.Button("auto", "Append", callback=self.appendRowButtonCallback)
        self.insertRowButton = vanilla.Button("auto", "Insert", callback=self.insertRowButtonCallback)
        self.removeRowButton = vanilla.Button("auto", "Remove", callback=self.removeRowButtonCallback)
        self.moveRowButton = vanilla.Button("auto", "Move", callback=self.moveRowButtonCallback)
        self.showRowButton = vanilla.Button("auto", "Show/Hide", callback=self.showRowButtonCallback)

        rows = [
            [
                vanilla.TextBox("auto", "Column:"), self.appendColumnButton, self.insertColumnButton, self.removeColumnButton, self.moveColumnButton, self.showColumnButton
            ],
            [
                vanilla.TextBox("auto", "Row:"), self.appendRowButton, self.insertRowButton, self.removeRowButton, self.moveRowButton, self.showRowButton
            ]
        ]
        columnDescriptions = [dict(width=60, columnPlacement="trailing")]
        for i in range(len(rows[0]) - 1):
            columnDescriptions.append({})

        self.w.line = vanilla.HorizontalLine("auto")
        self.w.testButtonGrid = vanilla.GridView(
            "auto",
            rows,
            columnDescriptions=columnDescriptions,
            columnWidth=100,
            columnSpacing=5,
            columnPlacement="fill",
            rowHeight=25
        )

        rules = [
            "H:|-margin-[gridView]-margin-|",
            "H:|[line]|",
            "H:|-margin-[testButtonGrid]-margin-|",
            "V:|"
               "-margin-"
               "[gridView]"
               "-margin-"
               "[line]"
               "-margin-"
               "[testButtonGrid]"
               "-margin-"
              "|"
        ]
        metrics = dict(
            margin=15
        )
        self.w.addAutoPosSizeRules(rules, metrics)
        self.w.open()

    def editTextCallback(self, sender):
        print("editTextCallback:", sender.get())

    # Column Tests

    def _makeRows(self):
        columnIndex = self.w.gridView.getColumnCount()
        rowCount = self.w.gridView.getRowCount()
        rows = [
            vanilla.TextBox(
                "auto",
                "C-{columnIndex}-{rowIndex}".format(
                    columnIndex=columnIndex,
                    rowIndex=rowIndex
                )
            )
            for rowIndex in range(rowCount)
        ]
        return rows

    def appendColumnButtonCallback(self, sender):
        rows = self._makeRows()
        self.w.gridView.appendColumn(rows, columnWidth=50)

    def insertColumnButtonCallback(self, sender):
        rows = self._makeRows()
        self.w.gridView.insertColumn(self.columnCount, rows, columnWidth=50)

    def removeColumnButtonCallback(self, sender):
        if self.w.gridView.getColumnCount() <= self.columnCount:
            # not a limitation of GridView, just a limitation of this test
            print("can't delete one of the main columns")
        else:
            self.w.gridView.removeColumn(self.columnCount)

    def moveColumnButtonCallback(self, sender):
        self.w.gridView.moveColumn(0, 1)

    def showColumnButtonCallback(self, sender):
        visible = self.w.gridView.columnIsVisible(1)
        self.w.gridView.showColumn(1, not visible)

    # Row Tests

    def _makeRow(self):
        columnCount = self.w.gridView.getColumnCount()
        rowIndex = self.w.gridView.getRowCount()
        cells = [
            vanilla.TextBox(
                "auto",
                "R-{columnIndex}-{rowIndex}".format(
                    columnIndex=columnIndex,
                    rowIndex=rowIndex
                )
            )
            for columnIndex in range(columnCount)
        ]
        return cells

    def appendRowButtonCallback(self, sender):
        row = self._makeRow()
        self.w.gridView.appendRow(row)

    def insertRowButtonCallback(self, sender):
        row = self._makeRow()
        self.w.gridView.insertRow(self.rowCount, row)

    def removeRowButtonCallback(self, sender):
        if self.w.gridView.getRowCount() <= self.rowCount:
            # not a limitation of GridView, just a limitation of this test
            print("can't delete one of the main rows")
        else:
            self.w.gridView.removeRow(self.rowCount)

    def moveRowButtonCallback(self, sender):
        self.w.gridView.moveRow(0, 1)

    def showRowButtonCallback(self, sender):
        visible = self.w.gridView.rowIsVisible(1)
        self.w.gridView.showRow(1, not visible)


if __name__ == "__main__":
    import vanilla
    from vanilla.test.testTools import executeVanillaTest
    executeVanillaTest(TestGridView)


import pprint
import AppKit
import vanilla

class Test:

    def __init__(self):
        self.w = vanilla.Window((300, 400))

        items = []
        for i in range(5):
            items.append(vanilla.List2GroupRow(f"Group {i + 1}"))
            items += list("123456789")

        self.w.list = vanilla.List2(
            "auto",
            items=items,
            allowsGroupRows=True,
            floatsGroupRows=True,
            allowsSorting=False,
            groupRowCellClass=CustomCell
        )
        rules = [
            "H:|[list]|",
            "V:|[list]|"
        ]
        self.w.addAutoPosSizeRules(rules)
        self.w.open()


class CustomCell(vanilla.Box):

    def __init__(self):
        super().__init__(
            (0, 0, 0, 0),
            fillColor=AppKit.NSColor.colorWithCalibratedRed_green_blue_alpha_(
                1, 0, 0, 0.25
            ),
            borderWidth=0,
            cornerRadius=10
        )
        self.textBox = vanilla.TextBox(
            (0, 0, 0, 0)
        )

    def set(self, value):
        self.textBox.set(value)

if __name__ == "__main__":
    from vanilla.test.testTools import executeVanillaTest
    executeVanillaTest(Test)



from AppKit import NSSwitchButton, NSShadowlessSquareBezelStyle, NSTextAlignmentLeft, NSNoCellMask
from vanilla.vanillaButton import Button
from vanilla.vanillaBase import osVersionCurrent, osVersion10_10, VanillaBaseObject


# In OS 10.0-10.X (tested up to OS 10.4) the small and mini check box
# buttons created in InterfaceBuilder can not be replicated programmatically.
# The text is not properly aligned with the check box button. To get around
# this we manually position the text next to the check box. Once Apple
# fixes this problem, we can revert back to the standard method for
# creating the control.

_doc = """
A standard check box.

.. image:: /_images/CheckBox.png

::

    from vanilla import Window, CheckBox

    class CheckBoxDemo:

        def __init__(self):
            self.w = Window((120, 40))
            self.w.checkBox = CheckBox((10, 10, -10, 20), "A CheckBox",
                               callback=self.checkBoxCallback, value=True)
            self.w.open()

        def checkBoxCallback(self, sender):
            print("check box state change!", sender.get())

    CheckBoxDemo()

**posSize** Tuple of form (left, top, width, height) or *"auto"* representing
the position and size of the check box. The size of the check box should match
the appropriate value for the given *sizeStyle*.

+-------------------------+
| **Standard Dimensions** |
+---------+---+-----------+
| Regular | H | 22        |
+---------+---+-----------+
| Small   | H | 18        |
+---------+---+-----------+
| Mini    | H | 10        |
+---------+---+-----------+

**title** The text to be displayed next to the check box. Pass *None* if
no title is desired.

**callback** The method to be called when the user changes the state of
the check box.

**value** A boolean representing the state of the check box.

**sizeStyle** A string representing the desired size style of the check box.
The options are:

+-----------+
| "regular" |
+-----------+
| "small"   |
+-----------+
| "mini"    |
+-----------+
"""


class _CheckBoxStandardBuild(Button):

    __doc__ = _doc

    nsButtonType = NSSwitchButton
    nsBezelStyle = None
    frameAdjustments = {
        "mini": (-4, -4, 6, 8),
        "small": (-3, -2, 5, 4),
        "regular": (-2, -2, 4, 4),
    }

    def __init__(self, posSize, title, callback=None, value=False, sizeStyle="regular"):
        super(_CheckBoxStandardBuild, self).__init__(posSize, title, callback=callback, sizeStyle=sizeStyle)
        self.set(value)

    def set(self, value):
        """
        Set the state of the check box.

        **value** A boolean representing the state of the check box.
        """
        self._nsObject.setState_(value)

    def get(self):
        """
        Get the state of the check box.
        """
        return self._nsObject.state()

    def toggle(self):
        """
        Toggle the state of the check box.

        If the check box is on, turn it off. If the check box is off, turn it on.
        """
        state = self.get()
        self.set(not state)


class _CheckBoxManualBuildButton(Button):

    nsButtonType = NSSwitchButton
    frameAdjustments = {
        "regular": (-2, -3, 4, 4),
        "small": (-3, -7, 5, 4),
        "mini": (-3, -11, 6, 8),
        }

    def set(self, value):
        self._nsObject.setState_(value)

    def get(self):
        return self._nsObject.state()

    def toggle(self):
        state = self.get()
        self.set(not state)


class _CheckBoxManualBuildTextButton(Button):

    nsBezelStyle = NSShadowlessSquareBezelStyle
    frameAdjustments = None

    def __init__(self, posSize, title, callback, sizeStyle):
        super(_CheckBoxManualBuildTextButton, self).__init__(posSize, title=title, callback=callback)
        self._nsObject.setBordered_(False)
        self._setSizeStyle(sizeStyle)
        self._nsObject.setAlignment_(NSTextAlignmentLeft)
        self._nsObject.cell().setHighlightsBy_(NSNoCellMask)


class _CheckBoxManualBuild(VanillaBaseObject):

    # both the container view and the check box will be adjusted.
    # this is necessary to create the appropriate buffer
    # and to handle the alignment.

    allFrameAdjustments = {
        "mini": (0, -4, 0, 8),
        "small": (0, -2, 0, 4),
        "regular": (0, -2, 0, 4),
        }

    def __init__(self, posSize, title, callback=None, value=False, sizeStyle="regular"):
        # This control is created by making a NSView that will contain
        # a NSButton set to check box mode and another NSButton set
        # to borderless mode to show the text. These buttons have their
        # callback set to this class, which then calls the callback
        # assigned to this class.
        self._setupView("NSView", posSize)

        self._callback = callback

        buttonSizes = {
                "mini": (10, 10),
                "small": (18, 18),
                "regular": (22, 22)
                }
        left, top, width, height = posSize

        self.frameAdjustments = self.allFrameAdjustments[sizeStyle]

        buttonWidth, buttonHeight = buttonSizes[sizeStyle]
        buttonLeft, buttonTop = self.frameAdjustments[:2]
        buttonLeft= abs(buttonLeft)
        buttonTop = abs(buttonTop)

        # adjust the position of the text button in relation to the check box
        textBoxPosSize = {
                # left, top, height
                "mini": (10, 4, 12),
                "small": (14, 4, 14),
                "regular": (16, 3, 17)
                }
        textBoxLeft, textBoxTop, textBoxHeight = textBoxPosSize[sizeStyle]
        textBoxWidth = 0

        self._checkBox = _CheckBoxManualBuildButton((0, 0, buttonWidth, buttonHeight), "", callback=self._buttonHit, sizeStyle=sizeStyle)
        self._checkBox.set(value)

        self._textButton = _CheckBoxManualBuildTextButton((textBoxLeft, textBoxTop, textBoxWidth, textBoxHeight), title=title, callback=self._buttonHit, sizeStyle=sizeStyle)

    def getNSButton(self):
        """
        Return the `NSButton`_ that this object wraps.

        *This is currently not implemented for CheckBox.*

        .. _NSButton: https://developer.apple.com/documentation/appkit/nsbutton?language=objc
        """
        # this is not possible since the control is built from parts
        raise NotImplementedError

    def _buttonHit(self, sender):
        # if the text box is the sender,
        # flip the state of the check box
        if sender == self._textButton:
            self._checkBox.toggle()
        if self._callback is not None:
            self._callback(self)

    def _breakCycles(self):
        self._callback = None
        self._checkBox._breakCycles()
        self._textButton._breakCycles()

    def enable(self, onOff):
        """
        Enable or disable the object. **onOff** should be a boolean.
        """
        self._checkBox.enable(onOff)
        self._textButton.enable(onOff)

    def setTitle(self, title):
        """
        Set the control title.

        **title** A string representing the title.
        """
        self._textButton.setTitle(title)

    def getTitle(self):
        """
        Get the control title.
        """
        return self._textButton.getTitle()

    def get(self):
        """
        Get the state of the check box.
        """
        return self._checkBox.get()

    def set(self, value):
        """
        Set the state of the check box.

        **value** A boolean representing the state of the check box.
        """
        self._checkBox.set(value)

    def toggle(self):
        """
        Toggle the state of the check box.

        If the check box is on, turn it off. If the check box is off, turn it on.
        """
        self._checkBox.toggle()


if osVersionCurrent >= osVersion10_10:
    class CheckBox(_CheckBoxStandardBuild):
        __doc__ = _doc
        pass
else:
    class CheckBox(_CheckBoxManualBuild):
        __doc__ = _doc
        pass



from AppKit import NSView, NSColor, NSBezierPath
from vanilla import VanillaBaseObject, Window, Button


class SimpleRectView(NSView):

    def isOpaque(self):
        return False

    def drawRect_(self, rect):
        NSColor.colorWithCalibratedWhite_alpha_(0.5, 0.5).set()
        p = NSBezierPath.bezierPathWithRect_(self.bounds())
        p.fill()


class SimpleRect(VanillaBaseObject):

    def __init__(self, posSize):
        self._nsObject = SimpleRectView.alloc().init()
        self._posSize = posSize
        self._setAutosizingFromPosSize(posSize)


class TestGeometry(object):

    def __init__(self):
        self.window = w = Window((400, 400), "Test Geometry", minSize=(100, 100))
        w.button = Button((10, 10, 120, 20), "call SetPosSize", callback=self.setFrameCallback)
        w.x = SimpleRect((10, 40, -120, -120))
        w.x.top = SimpleRect((20, 20, -20, 40))
        w.x.left = SimpleRect((40, 40, 40, -40))
        w.x.right = SimpleRect((-80, 40, 40, -40))
        w.x.lefttop = SimpleRect((10, 10, 50, 20))
        w.x.righttop = SimpleRect((-60, 10, 50, 20))
        w.x.bottom = SimpleRect((20, -60, -20, 40))
        w.x.leftbottom = SimpleRect((10, -30, 50, 20))
        w.x.rightbottom = SimpleRect((-60, -30, 50, 20))
        w.x.middle = SimpleRect((70, 70, -70, -70))
        w.x.middle.deep = SimpleRect((20, 10, -20, -10))
        w.extra = SimpleRect((-110, 10, -10, -120))
        w.y = SimpleRect((10, -110, -10, -20))
        w.y.z = SimpleRect((10, 10, -10, -10))
        w.y.z.q = SimpleRect((10, 10, -10, -10))
        w.open()

    def setFrameCallback(self, sender):
        l, t, r, b = self.window.x.getPosSize()
        l += 10
        t += 10
        r -= 10
        b -= 10
        self.window.x.setPosSize((l, t, r, b))



from AppKit import NSProgressIndicator, NSProgressIndicatorSpinningStyle
from vanilla.vanillaBase import VanillaBaseObject, _sizeStyleMap


class ProgressSpinner(VanillaBaseObject):

    """
    An animated, spinning progress indicator.

    .. image:: /_images/ProgressSpinner.png

    ::

        from vanilla import Window, ProgressSpinner

        class ProgressSpinnerDemo:

            def __init__(self):
                self.w = Window((80, 52))
                self.w.spinner = ProgressSpinner((24, 10, 32, 32),
                                        displayWhenStopped=True)
                self.w.spinner.start()
                self.w.open()

        ProgressSpinnerDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"*
    representing the position and size of the spinner. The size of the spinner
    should match the appropriate value for the given *sizeStyle*.

    +---------------------------+
    | **Standard Dimensions**   |
    +---------+---+----+---+----+
    | Regular | W | 32 | H | 32 |
    +---------+---+----+---+----+
    | Small   | W | 16 | H | 16 |
    +---------+---+----+---+----+

    **displayWhenStopped** Boolean representing if the spinner should be
    displayed when it is not spinning.

    **sizeStyle** A string representing the desired size style of the spinner.
    The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    """

    nsProgressIndicatorClass = NSProgressIndicator

    def __init__(self, posSize, displayWhenStopped=False, sizeStyle="regular"):
        self._setupView(self.nsProgressIndicatorClass, posSize)
        sizeStyle = _sizeStyleMap[sizeStyle]
        self._nsObject.setControlSize_(sizeStyle)
        self._nsObject.setStyle_(NSProgressIndicatorSpinningStyle)
        self._nsObject.setDisplayedWhenStopped_(displayWhenStopped)

    def getNSProgressIndicator(self):
        """
        Return the `NSProgressIndicator`_ that this object wraps.

        .. _NSProgressIndicator: https://developer.apple.com/documentation/appkit/nsprogressindicator?language=objc
        """
        return self._nsObject

    def start(self):
        """
        Start the animation.
        """
        self._nsObject.startAnimation_(None)

    def stop(self):
        """
        Stop the animation.
        """
        self._nsObject.stopAnimation_(None)



from warnings import warn
from vanilla.vanillaStackView import HorizontalStackView, VerticalStackView

class HorizontalStackGroup(HorizontalStackView):

    def __init__(self, posSize, *args, **kwargs):
        warn(
            DeprecationWarning("Use HorizontalStackView instead of HorizontalStackGroup.")
        )
        super().__init__(posSize, [], *args, **kwargs)

    def addView(self, *args, **kwargs):
        self.appendView(*args, **kwargs)


class VerticalStackGroup(VerticalStackView):

    def __init__(self, posSize, *args, **kwargs):
        warn(
            DeprecationWarning("Use VerticalStackView instead of VerticalStackGroup.")
        )
        super().__init__(posSize, [], *args, **kwargs)

    def addView(self, *args, **kwargs):
        self.appendView(*args, **kwargs)


from AppKit import NSTextField, NSAttributedString, NSTextAlignmentLeft, NSTextAlignmentRight, NSTextAlignmentCenter, NSTextAlignmentJustified, NSTextAlignmentNatural
from vanilla.vanillaBase import VanillaBaseControl


_textAlignmentMap = {
    "left": NSTextAlignmentLeft,
    "right": NSTextAlignmentRight,
    "center": NSTextAlignmentCenter,
    "justified": NSTextAlignmentJustified,
    "natural": NSTextAlignmentNatural,
}


class TextBox(VanillaBaseControl):

    """
    A rectangle containing static text.

    .. image:: /_images/TextBox.png

    ::

        from vanilla import Window, TextBox

        class TextBoxDemo:

             def __init__(self):
                 self.w = Window((100, 37))
                 self.w.textBox = TextBox((10, 10, -10, 17), "A TextBox")
                 self.w.open()

        TextBoxDemo()

    **posSize** Tuple of form *(left, top, width, height)* or *"auto"* representing
    the position and size of the text box.

    +-------------------------+
    | **Standard Dimensions** |
    +---------+---+-----------+
    | Regular | H | 17        |
    +---------+---+-----------+
    | Small   | H | 14        |
    +---------+---+-----------+
    | Mini    | H | 12        |
    +---------+---+-----------+

    **text** The text to be displayed in the text box. If the object is a
    `NSAttributedString`_, the attributes will be used for display.

    **alignment** A string representing the desired visual alignment of the
    text in the text box. The options are:

    +-------------+-----------------------------------------------------+
    | "left"      | Text is aligned left.                               |
    +-------------+-----------------------------------------------------+
    | "right"     | Text is aligned right.                              |
    +-------------+-----------------------------------------------------+
    | "center"    | Text is centered.                                   |
    +-------------+-----------------------------------------------------+
    | "justified" | Text is justified.                                  |
    +-------------+-----------------------------------------------------+
    | "natural"   | Follows the natural alignment of the text's script. |
    +-------------+-----------------------------------------------------+

    **selectable** Boolean representing if the text is selectable or not.

    **sizeStyle** A string representing the desired size style of the button.
    The options are:

    +-----------+
    | "regular" |
    +-----------+
    | "small"   |
    +-----------+
    | "mini"    |
    +-----------+

    .. _NSAttributedString: https://developer.apple.com/documentation/foundation/nsattributedstring?language=objc
    """

    nsTextFieldClass = NSTextField

    def __init__(self, posSize, text="", alignment="natural", selectable=False, sizeStyle="regular"):
        self._setupView(self.nsTextFieldClass, posSize)
        self._setSizeStyle(sizeStyle)
        if isinstance(text, NSAttributedString):
            self._nsObject.setAttributedStringValue_(text)
        else:
            self._nsObject.setStringValue_(text)
        self._nsObject.setDrawsBackground_(False)
        self._nsObject.setBezeled_(False)
        self._nsObject.setEditable_(False)
        self._nsObject.setSelectable_(selectable)
        self._nsObject.setAlignment_(_textAlignmentMap[alignment])

    def getNSTextField(self):
        """
        Return the `NSTextField`_ that this object wraps.

        .. _NSTextField: https://developer.apple.com/documentation/appkit/nstextfield?language=objc
        """
        return self._nsObject

    def get(self):
        """
        Get the contents of the text box.
        """
        return self._nsObject.stringValue()

    def set(self, value):
        """
        Set the contents of the text box.

        **value** A string representing the contents of the text box.
        """
        if isinstance(value, NSAttributedString):
            self._nsObject.setAttributedStringValue_(value)
        else:
            self._nsObject.setStringValue_(value)
